import uuid
from datetime import datetime

import pydantic

from . import ty
from .client.full_client import AccessTokenCredential, Client
from .models import MetricsCreating, MetricsLogging

SUPPORTED = [int, float, str]
MetricsValue = ty.Union[int, float, str, ty.BaseModel]


class Settings(ty.BaseSettings):
    pam_token: str = ""
    pam_host: str = ""
    job_id: ty.Optional[ty.UUID] = None
    execution_id: ty.Optional[ty.UUID] = None
    broker_id: ty.Optional[ty.UUID] = None

    def validate(self) -> bool:
        values = [
            self.pam_token,
            self.pam_host,
            self.job_id,
            self.execution_id,
            self.broker_id,
        ]
        return all(x for x in values)

    class Config:
        env_prefix = "PRODIGY_TEAMS_"


def get_pam_client(pam_host: str, pam_token: str) -> Client:
    client_base_url = pydantic.HttpUrl.build(scheme="https", host=pam_host)
    token = AccessTokenCredential(access_token=pam_token)
    return Client(client_base_url, token=token)


class Metrics:
    client: ty.Optional[Client]
    name: str
    id: ty.UUID
    type: ty.Type[MetricsValue]

    def __init__(self, name: str, metric_type: ty.Type[MetricsValue]) -> None:
        self._validate_type(metric_type)
        self.cfg = Settings()
        is_valid = self.cfg.validate()
        if is_valid:
            self.client = get_pam_client(self.cfg.pam_host, self.cfg.pam_token)
        else:
            self.client = None
        self.name = name
        self.id = uuid.uuid4()
        self.type = metric_type
        self.initialize()

    @property
    def type_schema(self) -> ty.Dict[str, ty.Any]:
        if isinstance(self.type, ty.BaseModel):
            return self.type.schema()
        return {"type": self.type.__name__}

    def initialize(self) -> None:
        if self.client is not None:
            assert self.cfg.job_id
            assert self.cfg.execution_id
            body = MetricsCreating(
                id=self.id,
                name=self.name,
                job_id=self.cfg.job_id,
                execution_id=self.cfg.execution_id,
                metrics_schema=self.type_schema,
            )
            self.client.for_cluster.init_metrics(body)
        else:  # TODO
            print("INIT METRICS", self.name, self.type_schema)

    def log(self, value: MetricsValue, final: bool = False) -> None:
        if not isinstance(value, self.type):
            raise ValueError(
                f"Can't log metric of type {type(value)}: doesn't match "
                f"specified metrics type ({self.type})"
            )
        data = value.json() if isinstance(value, ty.BaseModel) else value
        timestamp = datetime.now()
        if self.client is not None:
            body = MetricsLogging(
                timestamp=timestamp, id=self.id, value=data, final=final
            )
            self.client.for_cluster.log_metrics(body)
        else:  # TODO
            print("LOG METRICS", self.name, timestamp, data)

    def _validate_type(self, metric_type: ty.Type[MetricsValue]) -> None:
        if metric_type not in SUPPORTED and not isinstance(
            metric_type, type(ty.BaseModel)
        ):
            raise ValueError(
                f"Unsupported metrics type {metric_type}. The type can either "
                f"be one of {SUPPORTED} or a pydantic.BaseModel subclass"
            )



"""Generate a wheel for a prodigy_teams_recipes package,
including updating its meta file and requirements files."""
import json
import os
import re
import shlex
import shutil
import subprocess
from contextlib import ExitStack
from functools import cached_property
from pathlib import Path

from packaging.requirements import Requirement
from packaging.version import Version

from prodigy_teams.errors import RecipeBuildMetaFailed

from .. import ty
from ..build import (
    DirectorySource,
    RequirementSet,
    Venv,
    WheelSource,
    _get_venvs_cache_path,
    _make_tempdir,
)

RecipeInput = ty.Union[Requirement, WheelSource, DirectorySource]
ExtraWheelsInput = ty.Union[WheelSource, DirectorySource]


class RecipeBuilder:
    def __init__(
        self,
        src: RecipeInput,
        extras: ty.List[ExtraWheelsInput],
        wheelhouse: ty.Optional[Path] = None,
        build_dir: ty.Optional[Path] = None,
        cwd: ty.Optional[Path] = None,
        clear_cache: bool = False,
        target_python_version: ty.Optional[str] = None,
        target_platform: ty.Optional[str] = None,
        target_implementation: ty.Optional[str] = None,
        cache_pip_compile: bool = True,
    ):
        if clear_cache:
            venvs_cache = _get_venvs_cache_path()
            if venvs_cache.exists:
                shutil.rmtree(venvs_cache, ignore_errors=True)
        self._clear_cache = clear_cache
        self._cache_pip_compile = cache_pip_compile
        self._exit_stack = ExitStack()
        if build_dir is not None:
            self.build_dir = build_dir
        else:
            self.build_dir = self._exit_stack.enter_context(_make_tempdir())

        self.recipe_distribution_name = (
            src.name if isinstance(src, Requirement) else src.distribution_name
        )
        self.src = src
        self.extras = extras
        if wheelhouse is not None:
            self.wheelhouse = wheelhouse
        else:
            self.wheelhouse = self._exit_stack.enter_context(_make_tempdir())
        self._cwd = cwd
        self._upgrade_builder_venv = False
        self._target_python_version = target_python_version
        self._target_platform = target_platform
        self._target_implementation = target_implementation
        self._init_build_dir()

    def __enter__(self) -> "RecipeBuilder":
        return self

    def __exit__(self, _exc_type, _exc, _exc_tb):
        self._exit_stack.close()

    @cached_property
    def _builder_env_requirements(self) -> RequirementSet:
        return RequirementSet(
            ["setuptools==65.4.1", "wheel", "pip-tools>=6.13.0", "pip", "build>=0.10.0"]
        )

    @cached_property
    def _builder_venv(self) -> "Venv":
        with Venv.from_active(cwd=self._cwd) as venv:
            _builder_venv = Venv.cached(
                "__prodigy_teams__",
                build_venv=venv,
            )
            _builder_venv.install(
                *self._builder_env_requirements.to_lines(), upgrade=False
            )
        return _builder_venv

    @cached_property
    def recipe_version(self) -> Version:
        if isinstance(self.src, Requirement):
            version_spec = self.src.specifier
            if len(version_spec) == 1 and all(
                s.operator == "==" and s.version for s in version_spec
            ):
                return Version(next(iter(version_spec)).version)
            else:
                return self.prepared_wheelhouse[self.recipe_distribution_name].version
        elif isinstance(self.src, DirectorySource) and self.src.version is not None:
            return self.src.version
        else:
            return self.prepared_wheelhouse[self.recipe_distribution_name].version

    @cached_property
    def recipe_wheel(self) -> WheelSource:
        if isinstance(self.src, (Requirement, DirectorySource)):
            return self.prepared_wheelhouse[self.recipe_distribution_name]
        else:
            return self.src

    @cached_property
    def recipe_package_name(self) -> str:
        if isinstance(self.src, (Requirement, DirectorySource)):
            return self.prepared_wheelhouse[self.recipe_distribution_name].package_name
        else:
            return self.src.package_name

    @cached_property
    def resolved_input_requirements(self) -> RequirementSet:
        """
        Resolve the versions of the input packages.
        """
        requirements = set()
        for pkg in [self.src, *self.extras]:
            if isinstance(pkg, Requirement):
                requirements.add(pkg)
            elif pkg.version is not None:
                # requirements.add(Requirement(f"{pkg.distribution_name}=={pkg.version}"))
                requirements.add(Requirement(f"{pkg.distribution_name}=={pkg.version}"))
            else:
                # we need to build the wheel (metadata) to figure out the version
                built_wheels = self.prepared_wheelhouse
                wheel = built_wheels[pkg.distribution_name]
                requirements.add(
                    Requirement(f"{wheel.distribution_name}=={wheel.version}")
                )
        return RequirementSet(requirements)

    def _init_build_dir(self):
        local_requirements = self.build_dir / "recipe_requirements.txt"
        if self._cache_pip_compile:
            cached_venv = Venv._get_venv_cache_path(
                self.recipe_distribution_name, mkdir=False
            )
            cached_reqs = cached_venv / "requirements.txt"
            if cached_reqs.is_file():
                local_requirements.write_text(cached_reqs.read_text())

    @cached_property
    def compiled_venv_requirements(
        self,
    ) -> ty.Tuple[Path, Path, ty.List[str]]:
        local_reqs_file = self.build_dir / "recipe_requirements.txt"
        # cloud_reqs_file = self.build_dir / "cloud_requirements.txt"
        requirements = []
        if isinstance(self.src, Requirement):
            requirements.append(self.src)
        requirements.extend(self.prepared_wheelhouse.values())
        upgrade_packages = [r.name for r in self.resolved_input_requirements]
        compile_requirements(
            self._builder_venv,
            requirements=requirements,
            wheelhouse=self.wheelhouse,
            output_path=local_reqs_file,
            # upgrade_packages=upgrade_packages,
            upgrade=True,
            pip_tools_cache=self._get_pip_tools_cache(),
        )
        # Cross-compiling requirements is not supported by pip-compile
        # https://github.com/jazzband/pip-tools/pull/
        # compile_requirements(
        #     self._builder_venv,
        #     requirements=requirements,
        #     wheelhouse=self.wheelhouse,
        #     output_path=cloud_reqs_file,
        #     # upgrade_packages=upgrade_packages,
        #     upgrade=True,
        #     pip_tools_cache=self._get_pip_tools_cache(),
        #     python_version=self._target_python_version,
        #     python_platform=self._target_platform,
        #     python_implementation=self._target_implementation,
        # )
        return (
            local_reqs_file,
            local_reqs_file,
            upgrade_packages,
        )

    @cached_property
    def recipe_requirements_local(self) -> RequirementSet:
        reqs_file, _, _ = self.compiled_venv_requirements
        return RequirementSet.from_txt(reqs_file.read_text())

    @cached_property
    def recipes_venv(
        self,
    ) -> "Venv":
        recipes_venv = Venv.cached(
            self.recipe_distribution_name,
            self._builder_venv,
            requirements=self.recipe_requirements_local,
        )
        return recipes_venv

    @cached_property
    def prepared_wheelhouse(self) -> ty.Dict[str, WheelSource]:
        """
        Builds all local dependencies, and returns a mapping from distribution name to wheels.
        """
        wheels = []
        to_build = []
        for src in [self.src, *self.extras]:
            if isinstance(src, WheelSource):
                wheels.append(src)
            elif isinstance(src, DirectorySource):
                to_build.append(src)
            else:
                requirements = self.recipe_requirements_local
                requirements[self.recipe_distribution_name]
        wheels.extend(
            self._build_wheels(
                [p.path for p in to_build], self.wheelhouse, no_deps=True
            )
        )
        return {w.distribution_name: w for w in wheels}

    def _build_wheels(
        self,
        srcs: ty.List[Path],
        dest: Path,
        no_deps: bool = False,
        find_links: ty.List[Path] = [],
        cwd: ty.Optional[Path] = None,
    ) -> ty.List[WheelSource]:
        pip_wheel_args = [*[str(p.resolve()) for p in srcs], "-w", str(dest.absolute())]
        if no_deps:
            pip_wheel_args.append("--no-deps")
        if find_links:
            for dir in find_links:
                pip_wheel_args.extend(["-f", str(dir)])
        # SKIP_CYTHON is used here to ensure the prodigy wheel is platform independent.
        # - In the long run we need a better solution since this only works for optional
        #   native dependencies
        result = self._builder_venv.run_module(
            "pip",
            "wheel",
            *pip_wheel_args,
            cwd=cwd,
            env=dict(os.environ, SKIP_CYTHON="1"),
        )
        wheels = re.findall(r"(?<=filename=)\S+", result.stdout)
        return [WheelSource((dest / wheel).absolute()) for wheel in wheels]

    def _get_depcache(self) -> ty.List[Path]:
        pip_tools_cache = self._get_pip_tools_cache()
        return [p for p in pip_tools_cache.glob("depcache-cp*.json")]

    def _get_pip_tools_cache(self) -> Path:
        venv_cache = self._builder_venv._get_venv_cache_path(
            "__prodigy_teams__", mkdir=True
        )
        pip_tools_cache = venv_cache / "pip-tools-cache"
        pip_tools_cache.mkdir(exist_ok=True)
        return pip_tools_cache

    def _recipe_venv_path(self) -> Path:
        return self._builder_venv._get_venv_cache_path(
            self.recipe_distribution_name, mkdir=True
        )

    @cached_property
    def recipes_meta(
        self,
    ) -> ty.Dict[str, ty.Any]:
        recipes_venv = self.recipes_venv
        wheels = self.prepared_wheelhouse
        _, cloud_reqs_path, _ = self.compiled_venv_requirements
        cleaned_requirements = RequirementSet.from_txt(
            cloud_reqs_path
        ).replace_local_wheel_links(set(wheels.keys()))

        try:
            recipes_meta = json.loads(
                recipes_venv.run_module(
                    "prodigy_teams_recipes_sdk", "create-meta", self.recipe_package_name
                ).stdout
            )
        except subprocess.CalledProcessError as e:
            raise RecipeBuildMetaFailed(
                self.recipe_package_name, stdout=e.stdout, stderr=e.stderr
            )

        pkg_meta = self.recipe_wheel.pkginfo_metadata
        meta = {
            "name": self.recipe_wheel.distribution_name,
            "version": str(self.recipe_wheel.version),
            "description": pkg_meta.summary if pkg_meta is not None else None,
            "author": pkg_meta.author if pkg_meta is not None else None,
            "email": pkg_meta.author_email if pkg_meta is not None else None,
            "url": pkg_meta.home_page if pkg_meta is not None else None,
            "license": pkg_meta.license if pkg_meta is not None else None,
            "assets": {},
            "recipes": recipes_meta,
            "requirements": cleaned_requirements.to_lines(),
        }
        # Check it's json serializable
        _ = json.dumps(meta, indent=2)
        return meta


def compile_requirements(
    venv: Venv,
    requirements: ty.List[ty.Union[Requirement, WheelSource, str]],
    wheelhouse: Path,
    output_path: Path,
    pip_tools_cache: Path,
    upgrade_packages: ty.Iterable[str] = [],
    upgrade: bool = False,
    python_version: ty.Optional[str] = None,
    python_platform: ty.Optional[str] = None,
    python_implementation: ty.Optional[str] = None,
):
    # def collect_wheel_files(wheels: ty.List[WheelSource], wheelhouse: Path):
    #     """
    #     Get all of the local wheels we want to include so we can check
    #     if their requirements changed without a version change.
    #     """
    #     # TODO: this assumes the wheelhouse only includes the wheels we need
    #     # we should expand from the requirements in the initial wheels set
    #     # or send in the full set of wheels to be included
    #     all_wheels = {w.distribution_name: w for w in wheels}
    #     for wheel_path in wheelhouse.iterdir():
    #         if wheel_path.suffix == ".whl":
    #             wheel = WheelSource(wheel_path)
    #             if wheel.distribution_name not in all_wheels:
    #                 all_wheels[wheel.distribution_name] = wheel
    #             else:
    #                 # sanity check incase the wheelhouse contains multiple versions
    #                 assert (
    #                     all_wheels[wheel.distribution_name].version == wheel.version
    #                 ), f"Multiple versions of {wheel.distribution_name} found in wheelhouse (not implemented)"
    #     return all_wheels

    def refresh_pip_tools_cache(
        pip_tools_cache: Path, all_wheels: ty.Dict[str, WheelSource]
    ) -> Path:
        """
        Initializes the pip-compile cache, ensuring that the cache is not stale
        for any of the provided wheels.
        """
        for cache_file in pip_tools_cache.glob("depcache-cp*.json"):
            if not cache_file.is_file():
                continue

            cache = json.loads(cache_file.read_text())
            packages_to_refresh = {}

            for name, versions in cache.get("dependencies", {}).items():
                if (
                    name not in all_wheels
                    or str(all_wheels[name].version) not in versions
                ):
                    continue
                wheel = all_wheels[name]
                cached_reqs = versions.get(str(wheel.version), None)
                if set(Requirement(r) for r in cached_reqs) != wheel.requirements:
                    packages_to_refresh[(name, wheel.version)] = wheel.requirements

            if packages_to_refresh:
                for (name, version), requirements in packages_to_refresh.items():
                    package_cache = cache.setdefault("dependencies", {}).setdefault(
                        name, {}
                    )
                    if requirements is not None:
                        package_cache[str(version)] = [str(r) for r in requirements]
                    elif str(version) in package_cache:
                        del package_cache[str(version)]
                cache_file.write_text(json.dumps(cache))
        return pip_tools_cache

    wheels = []
    specs = []

    pip_args = []

    for req in requirements:
        if isinstance(req, WheelSource):
            wheels.append(req)
            wheel_path = req.path.resolve().as_uri()
            specs.append(f"{req.distribution_name} @ {wheel_path}")
        elif isinstance(req, str):
            specs.append(str(Requirement(req)))
        else:
            specs.append(str(req))

    if pip_tools_cache:
        refresh_pip_tools_cache(
            pip_tools_cache, {w.distribution_name: w for w in wheels}
        )
        cache_flag = ["--cache-dir", str(pip_tools_cache.resolve())]
    else:
        cache_flag = ["--rebuild"]

    upgrade_packages = list(
        set(upgrade_packages).union([w.distribution_name for w in wheels])
    )
    upgrade_packages_args = []
    if upgrade:
        upgrade_packages_args = ["--upgrade"]
    else:
        for to_upgrade in upgrade_packages:
            upgrade_packages_args.extend(["-P", to_upgrade])

    if python_version is not None:
        pip_args.extend(["--python-version", python_version])
    if python_platform is not None:
        pip_args.extend(["--platform", python_platform])
    if python_implementation is not None:
        pip_args.extend(["--implementation", python_implementation])

    return venv.run_module(
        "piptools",
        "compile",
        "-",  # read specs from stdin
        "--output-file",
        str(output_path.resolve()),
        "-f",
        str(wheelhouse.absolute()),
        "--no-emit-find-links",
        *cache_flag,
        *(["--pip-args", shlex.join(pip_args)] if pip_args else []),
        *upgrade_packages_args,
        input="\n".join(specs),
    )



# flake8: noqa
# generated by datamodel-codegen:
#   filename:  openapi.json

from __future__ import annotations

from datetime import datetime
from enum import Enum
from typing import Any, Dict, List, Optional, Union
from uuid import UUID

from pydantic import BaseModel, EmailStr, Field


class ActionDeleting(BaseModel):
    id: UUID = Field(..., title="Id")


class ActionState(Enum):
    NOT_FOUND = "NOT_FOUND"
    CREATED = "CREATED"
    PREPARING = "PREPARING"
    WITH_PROBLEMS = "WITH_PROBLEMS"
    STARTING = "STARTING"
    STARTED = "STARTED"
    NOT_STARTED = "NOT_STARTED"
    STOPPING = "STOPPING"
    STOPPED = "STOPPED"
    DELETING = "DELETING"
    PENDING = "PENDING"
    FAILED = "FAILED"
    NOMAD_ERROR = "NOMAD_ERROR"
    NOMAD_JOB_ERROR = "NOMAD_JOB_ERROR"
    OBTAINING_MODELS_VENV = "OBTAINING_MODELS_VENV"
    INITIALIZING_VENV = "INITIALIZING_VENV"
    COMPLETE = "COMPLETE"
    CANCELLED = "CANCELLED"
    EXECUTION_NOT_FOUND = "EXECUTION_NOT_FOUND"
    EXECUTION_RESULT_ERROR = "EXECUTION_RESULT_ERROR"
    NO_RESULTS_PATH_VALID = "NO_RESULTS_PATH_VALID"
    NO_EXECUTION = "NO_EXECUTION"


class ActionStateReturning(BaseModel):
    updated: datetime = Field(..., title="Updated")
    id: UUID = Field(..., title="Id")
    state: ActionState
    is_running: bool = Field(..., title="Is Running")
    is_startable: bool = Field(..., title="Is Startable")
    is_stoppable: bool = Field(..., title="Is Stoppable")


class ActionStats(BaseModel):
    executions: int = Field(..., title="Executions")
    last_run: Optional[datetime] = Field(None, title="Last Run")


class ActionUpdating(BaseModel):
    id: UUID = Field(..., title="Id")
    name: Optional[str] = Field(None, title="Name")


class AllocClientStatus(Enum):
    pending = "pending"
    running = "running"
    complete = "complete"
    failed = "failed"
    lost = "lost"
    unknown = "unknown"


class AllocDesiredStatus(Enum):
    run = "run"
    stop = "stop"
    evict = "evict"


class AssetCreating(BaseModel):
    broker_id: UUID = Field(..., title="Broker Id")
    name: str = Field(..., title="Name")
    version: str = Field(..., title="Version")
    path: str = Field(..., title="Path")
    kind: str = Field(..., title="Kind")
    meta: Dict[str, Any] = Field(..., title="Meta")


class AssetDeleting(BaseModel):
    id: UUID = Field(..., title="Id")


class AssetListingLatest(BaseModel):
    org_id: Optional[UUID] = Field(None, title="Org Id")
    broker_id: Optional[UUID] = Field(None, title="Broker Id")


class AssetReading(BaseModel):
    id: Optional[UUID] = Field(None, title="Id")
    broker_id: Optional[UUID] = Field(None, title="Broker Id")
    name: Optional[str] = Field(None, title="Name")
    version: Optional[str] = Field(None, title="Version")
    kind: Optional[str] = Field(None, title="Kind")
    path: Optional[str] = Field(None, title="Path")


class AssetReadingLatest(BaseModel):
    name: str = Field(..., title="Name")
    broker_id: UUID = Field(..., title="Broker Id")


class AssetSummary(BaseModel):
    id: UUID = Field(..., title="Id")
    created: datetime = Field(..., title="Created")
    updated: datetime = Field(..., title="Updated")
    broker_id: UUID = Field(..., title="Broker Id")
    name: str = Field(..., title="Name")
    version: str = Field(..., title="Version")
    kind: str = Field(..., title="Kind")
    path: str = Field(..., title="Path")
    meta: Dict[str, Any] = Field(..., title="Meta")
    num_used_by: int = Field(..., title="Num Used By")


class AssetUpdating(BaseModel):
    id: UUID = Field(..., title="Id")
    name: Optional[str] = Field(None, title="Name")
    version: Optional[str] = Field(None, title="Version")
    kind: Optional[str] = Field(None, title="Kind")
    path: Optional[str] = Field(None, title="Path")
    meta: Optional[Dict[str, Any]] = Field(None, title="Meta")


class BodyApiV1OauthTokenApiV1OauthTokenPost(BaseModel):
    grant_type: Optional[str] = Field(
        "client_credentials",
        description="OAuth2 Grant Type",
        regex="client_credentials",
        title="Grant Type",
    )
    scope: Optional[str] = Field("", description="OAuth2 Scope", title="Scope")
    client_id: Optional[str] = Field(
        None, description="Prodigy Teams Broker Client ID", title="Client Id"
    )
    client_secret: Optional[str] = Field(
        None, description="Prodigy Teams Broker Client Secret", title="Client Secret"
    )


class BrokerAuthenticated(BaseModel):
    access_token: str = Field(..., title="Access Token")
    expires_in: int = Field(..., title="Expires In")
    token_type: Optional[str] = Field("bearer", title="Token Type")


class BrokerDeleting(BaseModel):
    id: UUID = Field(..., title="Id")


class BrokerPathCreating(BaseModel):
    broker_id: UUID = Field(..., title="Broker Id")
    name: str = Field(..., title="Name")
    path: str = Field(..., title="Path")


class BrokerPathDeleting(BaseModel):
    id: UUID = Field(..., title="Id")


class BrokerPathDetail(BaseModel):
    id: UUID = Field(..., title="Id")
    created: datetime = Field(..., title="Created")
    updated: datetime = Field(..., title="Updated")
    broker_id: UUID = Field(..., title="Broker Id")
    name: str = Field(..., title="Name")
    path: str = Field(..., title="Path")


class BrokerPathReading(BaseModel):
    id: Optional[UUID] = Field(None, title="Id")
    broker_id: Optional[UUID] = Field(None, title="Broker Id")
    name: Optional[str] = Field(None, title="Name")
    path: Optional[str] = Field(None, title="Path")


class BrokerPathSummary(BaseModel):
    id: UUID = Field(..., title="Id")
    created: datetime = Field(..., title="Created")
    updated: datetime = Field(..., title="Updated")
    broker_id: UUID = Field(..., title="Broker Id")
    name: str = Field(..., title="Name")
    path: str = Field(..., title="Path")


class BrokerPathUpdating(BaseModel):
    id: UUID = Field(..., title="Id")
    name: str = Field(..., title="Name")
    path: str = Field(..., title="Path")


class BrokerReading(BaseModel):
    id: Optional[UUID] = Field(None, title="Id")
    org_id: Optional[UUID] = Field(None, title="Org Id")
    address: Optional[str] = Field(None, title="Address")
    state: Optional[str] = Field(None, title="State")
    region: Optional[str] = Field(None, title="Region")
    zone: Optional[str] = Field(None, title="Zone")
    cloud_provider: Optional[str] = Field(None, title="Cloud Provider")
    cloud_account: Optional[str] = Field(None, title="Cloud Account")
    cloud_region: Optional[str] = Field(None, title="Cloud Region")


class BrokerRegistering(BaseModel):
    address: str = Field(..., title="Address")
    state: str = Field(..., title="State")


class BrokerRegistrationCreating(BaseModel):
    name: str = Field(..., title="Name")
    cloud_provider: str = Field(..., title="Cloud Provider")
    cloud_account: Optional[str] = Field(None, title="Cloud Account")
    cloud_region: Optional[str] = Field(None, title="Cloud Region")


class BrokerRegistrationSummary(BaseModel):
    id: UUID = Field(..., title="Id")
    created: datetime = Field(..., title="Created")
    updated: datetime = Field(..., title="Updated")
    org_id: UUID = Field(..., title="Org Id")
    name: str = Field(..., title="Name")
    registration_token: UUID = Field(..., title="Registration Token")
    claimed: bool = Field(..., title="Claimed")
    expiration: datetime = Field(..., title="Expiration")


class BrokerSummary(BaseModel):
    id: UUID = Field(..., title="Id")
    created: datetime = Field(..., title="Created")
    updated: datetime = Field(..., title="Updated")
    org_id: UUID = Field(..., title="Org Id")
    name: str = Field(..., title="Name")
    address: str = Field(..., title="Address")
    state: str = Field(..., title="State")
    cloud_provider: str = Field(..., title="Cloud Provider")
    cloud_account: Optional[str] = Field(None, title="Cloud Account")
    cloud_region: Optional[str] = Field(None, title="Cloud Region")
    client_id: UUID = Field(..., title="Client Id")
    client_secret: str = Field(..., title="Client Secret")


class BrokerUpdating(BaseModel):
    id: UUID = Field(..., title="Id")
    name: Optional[str] = Field(None, title="Name")
    address: Optional[str] = Field(None, title="Address")


class CackQuerying(BaseModel):
    source: str = Field(..., title="Source")
    required: bool = Field(..., title="Required")
    fields: Optional[List[str]] = Field(None, title="Fields")
    filters: Optional[Dict[str, Any]] = Field({}, title="Filters")
    predicates: Optional[List[Dict[str, Any]]] = Field([], title="Predicates")
    relations: Optional[List[CackQuerying]] = Field([], title="Relations")
    key: Optional[str] = Field("", title="Key")
    return_list: Optional[bool] = Field(None, title="Return List")
    slice_start: Optional[int] = Field(None, title="Slice Start")
    slice_stop: Optional[int] = Field(None, title="Slice Stop")
    slice_step: Optional[int] = Field(None, title="Slice Step")


class Constraint(BaseModel):
    LTarget: str = Field(..., title="Ltarget")
    Operand: str = Field(..., title="Operand")
    RTarget: str = Field(..., title="Rtarget")


class DatasetCreating(BaseModel):
    name: str = Field(..., title="Name")
    broker_id: UUID = Field(..., title="Broker Id")
    kind: str = Field(..., title="Kind")


class DatasetDeleting(BaseModel):
    id: UUID = Field(..., title="Id")


class DatasetReading(BaseModel):
    id: Optional[UUID] = Field(None, title="Id")
    name: Optional[str] = Field(None, title="Name")
    kind: Optional[str] = Field(None, title="Kind")
    broker_id: Optional[UUID] = Field(None, title="Broker Id")


class DatasetSummary(BaseModel):
    id: UUID = Field(..., title="Id")
    created: datetime = Field(..., title="Created")
    updated: datetime = Field(..., title="Updated")
    name: str = Field(..., title="Name")
    broker_id: UUID = Field(..., title="Broker Id")
    kind: str = Field(..., title="Kind")
    num_used_by: int = Field(..., title="Num Used By")


class DatasetUpdating(BaseModel):
    id: UUID = Field(..., title="Id")
    name: Optional[str] = Field(None, title="Name")
    kind: Optional[str] = Field(None, title="Kind")


class EmailReturning(BaseModel):
    email: str = Field(..., title="Email")
    status: str = Field(..., title="Status")
    reject_reason: Optional[str] = Field(None, title="Reject Reason")
    queued_reason: Optional[str] = Field(None, title="Queued Reason")


class Empty(BaseModel):
    pass


class EmptyResponse(BaseModel):
    pass


class GroupAuthorizing(BaseModel):
    id: UUID = Field(..., title="Id")
    role: str = Field(..., title="Role")
    resource_type: str = Field(..., title="Resource Type")
    resource_id: UUID = Field(..., title="Resource Id")


class GroupCreating(BaseModel):
    name: str = Field(..., title="Name")
    org_id: UUID = Field(..., title="Org Id")


class GroupDeleting(BaseModel):
    id: UUID = Field(..., title="Id")


class GroupPopulating(BaseModel):
    id: UUID = Field(..., title="Id")
    user_id: UUID = Field(..., title="User Id")


class GroupReading(BaseModel):
    id: Optional[UUID] = Field(None, title="Id")
    name: Optional[str] = Field(None, title="Name")
    org_id: Optional[UUID] = Field(None, title="Org Id")


class GroupSummary(BaseModel):
    id: UUID = Field(..., title="Id")
    created: datetime = Field(..., title="Created")
    updated: datetime = Field(..., title="Updated")
    name: str = Field(..., title="Name")
    org_id: UUID = Field(..., title="Org Id")


class GroupUpdating(BaseModel):
    id: UUID = Field(..., title="Id")
    name: Optional[str] = Field(None, title="Name")


class InvitationChecked(BaseModel):
    id: UUID = Field(..., title="Id")
    created: datetime = Field(..., title="Created")
    updated: datetime = Field(..., title="Updated")


class InvitationChecking(BaseModel):
    id: UUID = Field(..., title="Id")
    email: Optional[str] = Field(None, title="Email")


class InvitationCreating(BaseModel):
    org_id: Optional[UUID] = Field(None, title="Org Id")
    email: str = Field(..., title="Email")
    vars: Optional[Dict[str, Any]] = Field({}, title="Vars")


class InvitationDeleting(BaseModel):
    id: UUID = Field(..., title="Id")


class InvitationDetail(BaseModel):
    id: UUID = Field(..., title="Id")
    created: datetime = Field(..., title="Created")
    updated: datetime = Field(..., title="Updated")
    org_id: UUID = Field(..., title="Org Id")
    email: str = Field(..., title="Email")
    claimed: bool = Field(..., title="Claimed")
    vars: Dict[str, Any] = Field(..., title="Vars")


class InvitationReading(BaseModel):
    id: Optional[UUID] = Field(None, title="Id")
    email: Optional[str] = Field(None, title="Email")
    claimed: Optional[bool] = Field(None, title="Claimed")


class InvitationSending(BaseModel):
    id: UUID = Field(..., title="Id")


class InvitationSummary(BaseModel):
    id: UUID = Field(..., title="Id")
    created: datetime = Field(..., title="Created")
    updated: datetime = Field(..., title="Updated")
    org_id: UUID = Field(..., title="Org Id")
    email: str = Field(..., title="Email")
    claimed: bool = Field(..., title="Claimed")
    vars: Dict[str, Any] = Field(..., title="Vars")


class InvitationUpdating(BaseModel):
    id: UUID = Field(..., title="Id")
    claimed: Optional[bool] = Field(None, title="Claimed")
    vars: Optional[Dict[str, Any]] = Field(None, title="Vars")


class InviteTaskUser(BaseModel):
    task_id: str = Field(..., title="Task Id")
    user_id: str = Field(..., title="User Id")
    notes: Optional[str] = Field(None, title="Notes")


class JobStatus(Enum):
    pending = "pending"
    running = "running"
    dead = "dead"


class JobType(Enum):
    task = "task"
    action = "action"


class LoginBody(BaseModel):
    org_id: Optional[UUID] = Field(None, title="Org Id")
    invitation_id: Optional[UUID] = Field(None, title="Invitation Id")


class MetricsCreating(BaseModel):
    id: UUID = Field(..., title="Id")
    name: str = Field(..., title="Name")
    job_id: UUID = Field(..., title="Job Id")
    execution_id: UUID = Field(..., title="Execution Id")
    metrics_schema: Dict[str, Any] = Field(..., title="Metrics Schema")


class MetricsLogging(BaseModel):
    timestamp: datetime = Field(..., title="Timestamp")
    id: UUID = Field(..., title="Id")
    value: Optional[Any] = Field(None, title="Value")
    final: bool = Field(..., title="Final")


class NomadAttemptUpdating(BaseModel):
    id: UUID = Field(..., title="Id")
    pam_error: Optional[Dict[str, Any]] = Field(None, title="Pam Error")
    validation_error: Optional[Dict[str, Any]] = Field(None, title="Validation Error")
    nomad_error: Optional[Dict[str, Any]] = Field(None, title="Nomad Error")
    evaluation_id: Optional[UUID] = Field(None, title="Evaluation Id")
    eval_create_index: Optional[int] = Field(None, title="Eval Create Index")
    job_modify_index: Optional[int] = Field(None, title="Job Modify Index")


class NomadEventCreating(BaseModel):
    broker_id: UUID = Field(..., title="Broker Id")
    index: int = Field(..., title="Index")
    topic: str = Field(..., title="Topic")
    type: str = Field(..., title="Type")
    key: str = Field(..., title="Key")
    payload: Dict[str, Any] = Field(..., title="Payload")
    filter_keys: Optional[List[str]] = Field([], title="Filter Keys")
    namespace: Optional[str] = Field("", title="Namespace")


class NomadEventDetail(BaseModel):
    id: Optional[UUID] = Field(None, title="Id")
    created: datetime = Field(..., title="Created")
    updated: datetime = Field(..., title="Updated")
    broker_id: Optional[UUID] = Field(None, title="Broker Id")
    action_id: Optional[UUID] = Field(None, title="Action Id")
    task_id: Optional[UUID] = Field(None, title="Task Id")
    index: Optional[int] = Field(None, title="Index")
    topic: Optional[str] = Field(None, title="Topic")
    type: Optional[str] = Field(None, title="Type")
    key: Optional[str] = Field(None, title="Key")
    namespace: Optional[str] = Field(None, title="Namespace")
    payload: Dict[str, Any] = Field(..., title="Payload")


class NomadEventGetIndex(BaseModel):
    latest_index: int = Field(..., title="Latest Index")


class NomadEventIndex(BaseModel):
    index: int = Field(..., title="Index")


class NomadEventReading(BaseModel):
    id: Optional[UUID] = Field(None, title="Id")
    broker_id: Optional[UUID] = Field(None, title="Broker Id")
    action_id: Optional[UUID] = Field(None, title="Action Id")
    task_id: Optional[UUID] = Field(None, title="Task Id")
    index: Optional[int] = Field(None, title="Index")
    topic: Optional[str] = Field(None, title="Topic")
    type: Optional[str] = Field(None, title="Type")
    key: Optional[str] = Field(None, title="Key")
    namespace: Optional[str] = Field(None, title="Namespace")


class NomadJobStub(BaseModel):
    CreateIndex: int = Field(..., title="Createindex")
    Datacenters: List[str] = Field(..., title="Datacenters")
    ID: UUID = Field(..., title="Id")
    JobModifyIndex: int = Field(..., title="Jobmodifyindex")
    ModifyIndex: int = Field(..., title="Modifyindex")
    Name: str = Field(..., title="Name")
    Namespace: Optional[str] = Field(None, title="Namespace")
    ParentID: Optional[UUID] = Field(None, title="Parentid")
    Periodic: Optional[bool] = Field(None, title="Periodic")
    Priority: Optional[int] = Field(None, title="Priority")
    Stable: Optional[bool] = Field(None, title="Stable")
    Status: JobStatus
    StatusDescription: str = Field(..., title="Statusdescription")
    Stop: bool = Field(..., title="Stop")
    SubmitTime: int = Field(..., title="Submittime")
    Type: Optional[str] = Field(None, title="Type")


class NomadJobsIds(BaseModel):
    jobs_ids: List[UUID] = Field(..., title="Jobs Ids")


class NotificationCreating(BaseModel):
    user_id: UUID = Field(..., title="User Id")
    type: str = Field(..., title="Type")
    message: str = Field(..., title="Message")
    is_read: bool = Field(..., title="Is Read")
    details: Dict[str, Union[str, int, float]] = Field(..., title="Details")


class NotificationDeleting(BaseModel):
    id: UUID = Field(..., title="Id")


class NotificationDetail(BaseModel):
    id: UUID = Field(..., title="Id")
    created: datetime = Field(..., title="Created")
    updated: datetime = Field(..., title="Updated")
    user_id: UUID = Field(..., title="User Id")
    type: str = Field(..., title="Type")
    message: str = Field(..., title="Message")
    is_read: bool = Field(..., title="Is Read")
    details: Dict[str, Any] = Field(..., title="Details")


class NotificationReading(BaseModel):
    id: Optional[UUID] = Field(None, title="Id")
    user_id: Optional[UUID] = Field(None, title="User Id")
    is_read: Optional[bool] = Field(None, title="Is Read")


class NotificationSummary(BaseModel):
    id: UUID = Field(..., title="Id")
    created: datetime = Field(..., title="Created")
    updated: datetime = Field(..., title="Updated")
    user_id: UUID = Field(..., title="User Id")
    type: str = Field(..., title="Type")
    message: str = Field(..., title="Message")
    is_read: bool = Field(..., title="Is Read")
    details: Dict[str, Any] = Field(..., title="Details")


class NotificationUpdating(BaseModel):
    id: UUID = Field(..., title="Id")
    is_read: Optional[bool] = Field(None, title="Is Read")
    message: Optional[str] = Field(None, title="Message")
    details: Optional[Dict[str, Union[str, int, float]]] = Field(None, title="Details")


class ObjectValidation(BaseModel):
    errors: Dict[str, List[str]] = Field(..., title="Errors")
    existing: Dict[str, List[str]] = Field(..., title="Existing")
    to_create: Dict[str, List[str]] = Field(..., title="To Create")


class OrgCreating(BaseModel):
    name: str = Field(..., title="Name")


class OrgDeleting(BaseModel):
    id: UUID = Field(..., title="Id")


class OrgDetail(BaseModel):
    id: UUID = Field(..., title="Id")
    created: datetime = Field(..., title="Created")
    updated: datetime = Field(..., title="Updated")
    name: str = Field(..., title="Name")


class OrgReading(BaseModel):
    id: Optional[UUID] = Field(None, title="Id")
    name: Optional[str] = Field(None, title="Name")


class OrgSummary(BaseModel):
    id: UUID = Field(..., title="Id")
    created: datetime = Field(..., title="Created")
    updated: datetime = Field(..., title="Updated")
    name: str = Field(..., title="Name")


class OrgUpdating(BaseModel):
    id: UUID = Field(..., title="Id")
    name: Optional[str] = Field(None, title="Name")


class PackageCreating(BaseModel):
    name: str = Field(..., title="Name")
    filename: str = Field(..., title="Filename")
    broker_id: UUID = Field(..., title="Broker Id")
    version: str = Field(..., title="Version")
    environment: str = Field(..., title="Environment")
    meta: Optional[Dict[str, Any]] = Field({}, title="Meta")


class PackageDeleting(BaseModel):
    id: UUID = Field(..., title="Id")
    force: Optional[bool] = Field(False, title="Force")


class PackageListingLatest(BaseModel):
    org_id: Optional[UUID] = Field(None, title="Org Id")
    broker_id: Optional[UUID] = Field(None, title="Broker Id")


class PackageReading(BaseModel):
    id: Optional[UUID] = Field(None, title="Id")
    name: Optional[str] = Field(None, title="Name")
    filename: Optional[str] = Field(None, title="Filename")
    org_id: Optional[UUID] = Field(None, title="Org Id")
    broker_id: Optional[UUID] = Field(None, title="Broker Id")
    version: Optional[str] = Field(None, title="Version")


class PackageReadingLatest(BaseModel):
    name: str = Field(..., title="Name")
    broker_id: UUID = Field(..., title="Broker Id")


class PackageSummary(BaseModel):
    id: UUID = Field(..., title="Id")
    created: datetime = Field(..., title="Created")
    updated: datetime = Field(..., title="Updated")
    name: str = Field(..., title="Name")
    version: str = Field(..., title="Version")
    environment: str = Field(..., title="Environment")
    num_used_by: int = Field(..., title="Num Used By")
    recipe_count: int = Field(..., title="Recipe Count")


class PackageUpdating(BaseModel):
    id: UUID = Field(..., title="Id")
    meta: Optional[Dict[str, Any]] = Field(None, title="Meta")


class PageAssetSummary(BaseModel):
    items: List[AssetSummary] = Field(..., title="Items")
    total: int = Field(..., ge=0, title="Total")
    page: int = Field(..., ge=1, title="Page")
    size: int = Field(..., ge=1, title="Size")


class PageBrokerPathSummary(BaseModel):
    items: List[BrokerPathSummary] = Field(..., title="Items")
    total: int = Field(..., ge=0, title="Total")
    page: int = Field(..., ge=1, title="Page")
    size: int = Field(..., ge=1, title="Size")


class PageBrokerSummary(BaseModel):
    items: List[BrokerSummary] = Field(..., title="Items")
    total: int = Field(..., ge=0, title="Total")
    page: int = Field(..., ge=1, title="Page")
    size: int = Field(..., ge=1, title="Size")


class PageDatasetSummary(BaseModel):
    items: List[DatasetSummary] = Field(..., title="Items")
    total: int = Field(..., ge=0, title="Total")
    page: int = Field(..., ge=1, title="Page")
    size: int = Field(..., ge=1, title="Size")


class PageGroupSummary(BaseModel):
    items: List[GroupSummary] = Field(..., title="Items")
    total: int = Field(..., ge=0, title="Total")
    page: int = Field(..., ge=1, title="Page")
    size: int = Field(..., ge=1, title="Size")


class PageInvitationSummary(BaseModel):
    items: List[InvitationSummary] = Field(..., title="Items")
    total: int = Field(..., ge=0, title="Total")
    page: int = Field(..., ge=1, title="Page")
    size: int = Field(..., ge=1, title="Size")


class PageNotificationSummary(BaseModel):
    items: List[NotificationSummary] = Field(..., title="Items")
    total: int = Field(..., ge=0, title="Total")
    page: int = Field(..., ge=1, title="Page")
    size: int = Field(..., ge=1, title="Size")


class PageOrgSummary(BaseModel):
    items: List[OrgSummary] = Field(..., title="Items")
    total: int = Field(..., ge=0, title="Total")
    page: int = Field(..., ge=1, title="Page")
    size: int = Field(..., ge=1, title="Size")


class PagePackageSummary(BaseModel):
    items: List[PackageSummary] = Field(..., title="Items")
    total: int = Field(..., ge=0, title="Total")
    page: int = Field(..., ge=1, title="Page")
    size: int = Field(..., ge=1, title="Size")


class PersonCreating(BaseModel):
    name: str = Field(..., title="Name")
    source: str = Field(..., title="Source")
    email: EmailStr = Field(..., title="Email")
    picture: str = Field(..., title="Picture")
    locale: str = Field(..., title="Locale")
    avatar_url: str = Field(..., title="Avatar Url")
    preferences: Dict[str, Any] = Field(..., title="Preferences")


class PersonDeleting(BaseModel):
    id: Optional[UUID] = Field(None, title="Id")
    email: Optional[str] = Field(None, title="Email")


class PersonDetail(BaseModel):
    id: UUID = Field(..., title="Id")
    created: datetime = Field(..., title="Created")
    updated: datetime = Field(..., title="Updated")
    name: str = Field(..., title="Name")
    source: str = Field(..., title="Source")
    email: str = Field(..., title="Email")
    picture: str = Field(..., title="Picture")
    avatar_url: str = Field(..., title="Avatar Url")
    locale: str = Field(..., title="Locale")
    verified: bool = Field(..., title="Verified")
    preferences: Dict[str, Any] = Field(..., title="Preferences")


class PersonReading(BaseModel):
    id: Optional[UUID] = Field(None, title="Id")
    name: Optional[str] = Field(None, title="Name")
    source: Optional[str] = Field(None, title="Source")
    email: Optional[str] = Field(None, title="Email")
    verified: Optional[bool] = Field(None, title="Verified")


class PersonSummary(BaseModel):
    id: UUID = Field(..., title="Id")
    created: datetime = Field(..., title="Created")
    updated: datetime = Field(..., title="Updated")
    name: str = Field(..., title="Name")
    source: str = Field(..., title="Source")
    email: str = Field(..., title="Email")
    picture: str = Field(..., title="Picture")
    avatar_url: str = Field(..., title="Avatar Url")
    locale: str = Field(..., title="Locale")
    verified: bool = Field(..., title="Verified")
    preferences: Dict[str, Any] = Field(..., title="Preferences")


class PersonUpdating(BaseModel):
    id: UUID = Field(..., title="Id")
    name: Optional[str] = Field(None, title="Name")
    source: Optional[str] = Field(None, title="Source")
    email: Optional[str] = Field(None, title="Email")
    picture: Optional[str] = Field(None, title="Picture")
    avatar_url: Optional[str] = Field(None, title="Avatar Url")
    locale: Optional[str] = Field(None, title="Locale")
    preferences: Optional[Dict[str, Any]] = Field(None, title="Preferences")
    verified: Optional[bool] = Field(None, title="Verified")
    last_push_time: Optional[datetime] = Field(None, title="Last Push Time")


class ProjectCreating(BaseModel):
    name: str = Field(..., min_length=3, title="Name")
    description: str = Field(..., title="Description")


class ProjectDeleting(BaseModel):
    id: UUID = Field(..., title="Id")


class ProjectDetail(BaseModel):
    id: UUID = Field(..., title="Id")
    created: datetime = Field(..., title="Created")
    updated: datetime = Field(..., title="Updated")
    org_id: UUID = Field(..., title="Org Id")
    name: str = Field(..., title="Name")
    description: str = Field(..., title="Description")


class ProjectReading(BaseModel):
    id: Optional[UUID] = Field(None, title="Id")
    name: Optional[str] = Field(None, title="Name")
    org_id: Optional[UUID] = Field(None, title="Org Id")


class ProjectStats(BaseModel):
    annotations: int = Field(..., title="Annotations")
    annotators: int = Field(..., title="Annotators")
    live_sessions: int = Field(..., title="Live Sessions")
    tasks_count: int = Field(..., title="Tasks Count")
    actions_count: int = Field(..., title="Actions Count")
    actions_executions: int = Field(..., title="Actions Executions")
    actions_failed: int = Field(..., title="Actions Failed")


class ProjectSummary(BaseModel):
    id: UUID = Field(..., title="Id")
    created: datetime = Field(..., title="Created")
    updated: datetime = Field(..., title="Updated")
    org_id: UUID = Field(..., title="Org Id")
    name: str = Field(..., title="Name")
    description: str = Field(..., title="Description")


class ProjectUpdating(BaseModel):
    id: UUID = Field(..., title="Id")
    name: Optional[str] = Field(None, title="Name")
    description: Optional[str] = Field(None, title="Description")


class RecipeDeleting(BaseModel):
    id: UUID = Field(..., title="Id")


class RecipeListingLatest(BaseModel):
    org_id: Optional[UUID] = Field(None, title="Org Id")
    broker_id: Optional[UUID] = Field(None, title="Broker Id")


class RecipePlanCreating(BaseModel):
    recipe_id: UUID = Field(..., title="Recipe Id")
    project_id: UUID = Field(..., title="Project Id")
    args: Dict[str, Any] = Field(..., title="Args")
    worker_class: Optional[str] = Field(None, title="Worker Class")


class RecipePlanDeleting(BaseModel):
    id: UUID = Field(..., title="Id")


class RecipePlanDetail(BaseModel):
    id: UUID = Field(..., title="Id")
    created: datetime = Field(..., title="Created")
    updated: datetime = Field(..., title="Updated")
    project_id: UUID = Field(..., title="Project Id")
    project_name: str = Field(..., title="Project Name")
    recipe_id: UUID = Field(..., title="Recipe Id")
    recipe_name: str = Field(..., title="Recipe Name")
    recipe_title: Optional[str] = Field(None, title="Recipe Title")
    recipe_description: Optional[str] = Field(None, title="Recipe Description")
    package_id: UUID = Field(..., title="Package Id")
    package_name: str = Field(..., title="Package Name")
    package_version: str = Field(..., title="Package Version")
    environment: str = Field(..., title="Environment")
    worker_class: Optional[str] = Field(None, title="Worker Class")
    args: Dict[str, Any] = Field(..., title="Args")
    command: List[str] = Field(..., title="Command")
    objects: Dict[str, Dict[str, Dict[str, Any]]] = Field(..., title="Objects")


class RecipePlanReading(BaseModel):
    id: Optional[UUID] = Field(None, title="Id")
    recipe_id: Optional[UUID] = Field(None, title="Recipe Id")
    project_id: Optional[UUID] = Field(None, title="Project Id")


class RecipePlanSummary(BaseModel):
    id: UUID = Field(..., title="Id")
    created: datetime = Field(..., title="Created")
    updated: datetime = Field(..., title="Updated")
    project_id: UUID = Field(..., title="Project Id")
    project_name: str = Field(..., title="Project Name")
    recipe_id: UUID = Field(..., title="Recipe Id")
    recipe_name: str = Field(..., title="Recipe Name")
    recipe_title: Optional[str] = Field(None, title="Recipe Title")
    recipe_description: Optional[str] = Field(None, title="Recipe Description")
    package_id: UUID = Field(..., title="Package Id")
    package_name: str = Field(..., title="Package Name")
    package_version: str = Field(..., title="Package Version")
    environment: str = Field(..., title="Environment")
    worker_class: Optional[str] = Field(None, title="Worker Class")


class RecipePlanUpdating(BaseModel):
    id: UUID = Field(..., title="Id")
    worker_class: Optional[str] = Field(None, title="Worker Class")


class RecipeReading(BaseModel):
    id: Optional[UUID] = Field(None, title="Id")
    org_id: Optional[UUID] = Field(None, title="Org Id")
    name: Optional[str] = Field(None, title="Name")
    package_id: Optional[UUID] = Field(None, title="Package Id")


class RecipeReadingLatest(BaseModel):
    name: str = Field(..., title="Name")
    broker_id: UUID = Field(..., title="Broker Id")


class RecipeSummary(BaseModel):
    id: UUID = Field(..., title="Id")
    created: datetime = Field(..., title="Created")
    updated: datetime = Field(..., title="Updated")
    org_id: UUID = Field(..., title="Org Id")
    package_id: UUID = Field(..., title="Package Id")
    name: str = Field(..., title="Name")
    title: Optional[str] = Field(None, title="Title")
    description: Optional[str] = Field(None, title="Description")
    is_action: bool = Field(..., title="Is Action")
    entry_point: Optional[str] = Field(None, title="Entry Point")
    num_used_by: int = Field(..., title="Num Used By")


class RecipeUpdating(BaseModel):
    id: UUID = Field(..., title="Id")
    title: str = Field(..., title="Title")


class ReportNomadJobResponse(BaseModel):
    missing: List[UUID] = Field(..., title="Missing")


class Resources(BaseModel):
    CPU: Optional[int] = Field(None, title="Cpu")
    Cores: Optional[int] = Field(None, title="Cores")
    MemoryMB: Optional[int] = Field(None, title="Memorymb")
    MemoryMaxMB: Optional[int] = Field(None, title="Memorymaxmb")
    DiskMB: Optional[int] = Field(None, title="Diskmb")
    IOPS: Optional[int] = Field(None, title="Iops")
    Networks: Optional[Dict[str, Any]] = Field(None, title="Networks")
    Devices: Optional[Dict[str, Any]] = Field(None, title="Devices")


class SecretCreating(BaseModel):
    name: str = Field(..., title="Name")
    path: str = Field(..., title="Path")
    broker_id: UUID = Field(..., title="Broker Id")


class SecretDeleting(BaseModel):
    id: UUID = Field(..., title="Id")


class SecretDetail(BaseModel):
    id: UUID = Field(..., title="Id")
    created: datetime = Field(..., title="Created")
    updated: datetime = Field(..., title="Updated")
    name: str = Field(..., title="Name")
    path: str = Field(..., title="Path")
    broker_id: UUID = Field(..., title="Broker Id")


class SecretReading(BaseModel):
    name: Optional[str] = Field(None, title="Name")
    path: Optional[str] = Field(None, title="Path")
    broker_id: Optional[UUID] = Field(None, title="Broker Id")


class SecretSummary(BaseModel):
    id: UUID = Field(..., title="Id")
    created: datetime = Field(..., title="Created")
    updated: datetime = Field(..., title="Updated")
    name: str = Field(..., title="Name")
    path: str = Field(..., title="Path")
    broker_id: UUID = Field(..., title="Broker Id")


class SecretUpdating(BaseModel):
    id: UUID = Field(..., title="Id")
    name: str = Field(..., title="Name")
    path: str = Field(..., title="Path")


class SessionCreating(BaseModel):
    user_id: UUID = Field(..., title="User Id")
    task_id: UUID = Field(..., title="Task Id")


class SessionDeleting(BaseModel):
    id: UUID = Field(..., title="Id")


class SessionDetail(BaseModel):
    id: UUID = Field(..., title="Id")
    created: datetime = Field(..., title="Created")
    updated: datetime = Field(..., title="Updated")
    user_id: UUID = Field(..., title="User Id")
    user_name: str = Field(..., title="User Name")
    user_avatar_url: str = Field(..., title="User Avatar Url")
    task_id: UUID = Field(..., title="Task Id")
    count: int = Field(..., title="Count")
    time_spent: int = Field(..., title="Time Spent")


class SessionReading(BaseModel):
    id: Optional[UUID] = Field(None, title="Id")
    user_id: Optional[UUID] = Field(None, title="User Id")
    task_id: Optional[UUID] = Field(None, title="Task Id")
    count: Optional[int] = Field(None, title="Count")
    time_spent: Optional[int] = Field(None, title="Time Spent")


class SessionSaving(BaseModel):
    id: UUID = Field(..., title="Id")
    count: Optional[int] = Field(None, title="Count")
    time_spent: Optional[int] = Field(None, title="Time Spent")


class SessionSummary(BaseModel):
    id: UUID = Field(..., title="Id")
    created: datetime = Field(..., title="Created")
    updated: datetime = Field(..., title="Updated")
    user_id: UUID = Field(..., title="User Id")
    user_name: str = Field(..., title="User Name")
    user_avatar_url: str = Field(..., title="User Avatar Url")
    task_id: UUID = Field(..., title="Task Id")
    count: int = Field(..., title="Count")
    time_spent: int = Field(..., title="Time Spent")


class SessionTokenReturning(BaseModel):
    session_id: UUID = Field(..., title="Session Id")
    user_id: UUID = Field(..., title="User Id")
    task_id: UUID = Field(..., title="Task Id")
    prodigy_token: Optional[str] = Field(None, title="Prodigy Token")


class SessionUpdating(BaseModel):
    id: UUID = Field(..., title="Id")
    count: Optional[int] = Field(None, title="Count")


class TaskAnnotator(BaseModel):
    id: UUID = Field(..., title="Id")
    created: datetime = Field(..., title="Created")
    updated: datetime = Field(..., title="Updated")
    name: str = Field(..., title="Name")
    org_id: UUID = Field(..., title="Org Id")
    org_name: str = Field(..., title="Org Name")
    person_id: UUID = Field(..., title="Person Id")
    email: str = Field(..., title="Email")
    avatar_url: str = Field(..., title="Avatar Url")
    locale: str = Field(..., title="Locale")
    is_org_admin: bool = Field(..., title="Is Org Admin")
    is_org_developer: bool = Field(..., title="Is Org Developer")
    groups: List[str] = Field(..., title="Groups")
    tags: List[str] = Field(..., title="Tags")
    annotations: int = Field(..., title="Annotations")
    live_sessions: int = Field(..., title="Live Sessions")
    sessions: int = Field(..., title="Sessions")
    hours: str = Field(..., title="Hours")
    is_added: bool = Field(..., title="Is Added")


class TaskCreating(BaseModel):
    name: str = Field(..., title="Name")
    project_id: UUID = Field(..., title="Project Id")
    plan: RecipePlanCreating


class TaskDeleting(BaseModel):
    id: UUID = Field(..., title="Id")


class TaskEventDetails(BaseModel):
    fails_task: Optional[bool] = Field(False, title="Fails Task")
    restart_reason: Optional[str] = Field(None, title="Restart Reason")
    setup_error: Optional[str] = Field(None, title="Setup Error")
    driver_error: Optional[str] = Field(None, title="Driver Error")
    exit_code: Optional[int] = Field(None, title="Exit Code")
    signal: Optional[int] = Field(None, title="Signal")
    kill_timeout: Optional[str] = Field(None, title="Kill Timeout")
    kill_error: Optional[str] = Field(None, title="Kill Error")
    kill_reason: Optional[str] = Field(None, title="Kill Reason")
    start_delay: Optional[int] = Field(None, title="Start Delay")
    download_error: Optional[str] = Field(None, title="Download Error")
    validation_error: Optional[str] = Field(None, title="Validation Error")
    disk_limit: Optional[int] = Field(None, title="Disk Limit")
    failed_sibling: Optional[str] = Field(None, title="Failed Sibling")
    vault_error: Optional[str] = Field(None, title="Vault Error")
    task_signal_reason: Optional[str] = Field(None, title="Task Signal Reason")
    task_signal: Optional[str] = Field(None, title="Task Signal")


class TaskEventType(Enum):
    Alloc_Unhealthy = "Alloc Unhealthy"
    Setup_Failure = "Setup Failure"
    Driver_Failure = "Driver Failure"
    Received = "Received"
    Failed_Validation = "Failed Validation"
    Started = "Started"
    Terminated = "Terminated"
    Killing = "Killing"
    Killed = "Killed"
    Restarting = "Restarting"
    Not_Restarting = "Not Restarting"
    Restart_Signaled = "Restart Signaled"
    Signaling = "Signaling"
    Downloading_Artifacts = "Downloading Artifacts"
    Failed_Artifact_Download = "Failed Artifact Download"
    Building_Task_Directory = "Building Task Directory"
    Task_Setup = "Task Setup"
    Disk_Resources_Exceeded = "Disk Resources Exceeded"
    Sibling_Task_Failed = "Sibling Task Failed"
    Driver = "Driver"
    Leader_Task_Dead = "Leader Task Dead"
    Main_Tasks_Dead = "Main Tasks Dead"
    Task_hook_failed = "Task hook failed"
    Task_hook_message = "Task hook message"
    Failed_Restoring_Task = "Failed Restoring Task"
    Plugin_became_unhealthy = "Plugin became unhealthy"
    Plugin_became_healthy = "Plugin became healthy"
    Reconnected = "Reconnected"
    Waiting_for_shutdown_delay = "Waiting for shutdown delay"


class TaskPopulating(BaseModel):
    id: UUID = Field(..., title="Id")
    user_id: UUID = Field(..., title="User Id")


class TaskState(Enum):
    NOT_FOUND = "NOT_FOUND"
    CREATED = "CREATED"
    PREPARING = "PREPARING"
    WITH_PROBLEMS = "WITH_PROBLEMS"
    STARTING = "STARTING"
    STARTED = "STARTED"
    NOT_STARTED = "NOT_STARTED"
    STOPPING = "STOPPING"
    STOPPED = "STOPPED"
    DELETING = "DELETING"
    PENDING = "PENDING"
    FAILED = "FAILED"
    NOMAD_ERROR = "NOMAD_ERROR"
    NOMAD_JOB_ERROR = "NOMAD_JOB_ERROR"
    OBTAINING_MODELS_VENV = "OBTAINING_MODELS_VENV"
    INITIALIZING_VENV = "INITIALIZING_VENV"


class TaskStateReturning(BaseModel):
    updated: datetime = Field(..., title="Updated")
    id: UUID = Field(..., title="Id")
    state: TaskState
    is_running: bool = Field(..., title="Is Running")
    is_startable: bool = Field(..., title="Is Startable")
    is_stoppable: bool = Field(..., title="Is Stoppable")


class TaskStats(BaseModel):
    annotations: int = Field(..., title="Annotations")
    annotators: int = Field(..., title="Annotators")
    live_sessions: int = Field(..., title="Live Sessions")
    sessions: int = Field(..., title="Sessions")
    hours: str = Field(..., title="Hours")


class TaskSummary(BaseModel):
    id: UUID = Field(..., title="Id")
    created: datetime = Field(..., title="Created")
    updated: datetime = Field(..., title="Updated")
    project_id: UUID = Field(..., title="Project Id")
    broker_id: UUID = Field(..., title="Broker Id")
    name: str = Field(..., title="Name")
    recipe_name: str = Field(..., title="Recipe Name")
    recipe_title: str = Field(..., title="Recipe Title")
    state: TaskState
    project_name: str = Field(..., title="Project Name")
    is_running: bool = Field(..., title="Is Running")
    is_startable: bool = Field(..., title="Is Startable")
    is_stoppable: bool = Field(..., title="Is Stoppable")
    error: Optional[str] = Field(None, title="Error")
    plan: RecipePlanSummary
    job_type: JobType
    cli_command: str = Field(..., title="Cli Command")


class TaskUpdating(BaseModel):
    id: UUID = Field(..., title="Id")
    name: Optional[str] = Field(None, title="Name")


class TypeSchema(BaseModel):
    name: str = Field(..., title="Name")
    args: Optional[List[Union[str, TypeSchema]]] = Field(None, title="Args")
    object_type: Optional[str] = Field(None, title="Object Type")


class UserAuthenticated(BaseModel):
    access_token: str = Field(..., title="Access Token")
    expires_in: int = Field(..., title="Expires In")
    token_type: Optional[str] = Field("bearer", title="Token Type")


class UserCreating(BaseModel):
    org_id: UUID = Field(..., title="Org Id")
    person_id: UUID = Field(..., title="Person Id")
    is_org_admin: bool = Field(..., title="Is Org Admin")
    is_org_developer: bool = Field(..., title="Is Org Developer")
    tags: Optional[List[str]] = Field([], title="Tags")


class UserDeleting(BaseModel):
    id: Optional[UUID] = Field(None, title="Id")
    org_id: Optional[UUID] = Field(None, title="Org Id")
    person_id: Optional[UUID] = Field(None, title="Person Id")


class UserDetail(BaseModel):
    id: UUID = Field(..., title="Id")
    created: datetime = Field(..., title="Created")
    updated: datetime = Field(..., title="Updated")
    name: str = Field(..., title="Name")
    org_id: UUID = Field(..., title="Org Id")
    org_name: str = Field(..., title="Org Name")
    person_id: UUID = Field(..., title="Person Id")
    email: str = Field(..., title="Email")
    avatar_url: str = Field(..., title="Avatar Url")
    locale: str = Field(..., title="Locale")
    is_org_admin: bool = Field(..., title="Is Org Admin")
    is_org_developer: bool = Field(..., title="Is Org Developer")
    groups: List[str] = Field(..., title="Groups")
    tags: List[str] = Field(..., title="Tags")


class UserReading(BaseModel):
    id: Optional[UUID] = Field(None, title="Id")
    org_id: Optional[UUID] = Field(None, title="Org Id")
    person_id: Optional[UUID] = Field(None, title="Person Id")
    is_org_admin: Optional[bool] = Field(None, title="Is Org Admin")
    is_org_developer: Optional[bool] = Field(None, title="Is Org Developer")


class UserSummary(BaseModel):
    id: UUID = Field(..., title="Id")
    created: datetime = Field(..., title="Created")
    updated: datetime = Field(..., title="Updated")
    name: str = Field(..., title="Name")
    org_id: UUID = Field(..., title="Org Id")
    org_name: str = Field(..., title="Org Name")
    person_id: UUID = Field(..., title="Person Id")
    email: str = Field(..., title="Email")
    avatar_url: str = Field(..., title="Avatar Url")
    locale: str = Field(..., title="Locale")
    is_org_admin: bool = Field(..., title="Is Org Admin")
    is_org_developer: bool = Field(..., title="Is Org Developer")


class UserUpdating(BaseModel):
    id: UUID = Field(..., title="Id")
    name: Optional[str] = Field(None, title="Name")
    email: Optional[str] = Field(None, title="Email")
    avatar_url: Optional[str] = Field(None, title="Avatar Url")
    locale: Optional[str] = Field(None, title="Locale")
    tags: Optional[List[str]] = Field([], title="Tags")
    is_org_admin: Optional[bool] = Field(None, title="Is Org Admin")
    is_org_developer: Optional[bool] = Field(None, title="Is Org Developer")


class ValidationError(BaseModel):
    loc: List[Union[str, int]] = Field(..., title="Location")
    msg: str = Field(..., title="Message")
    type: str = Field(..., title="Error Type")


class ValidationLevel(Enum):
    error = "error"
    warning = "warning"
    info = "info"
    success = "success"


class ValidationOp(Enum):
    ge = "ge"
    gt = "gt"
    le = "le"
    lt = "lt"
    eq = "eq"
    ne = "ne"
    re = "re"


class ValidationSchema(BaseModel):
    op: ValidationOp
    value: Union[int, float, str, bool] = Field(..., title="Value")
    message: str = Field(..., title="Message")
    level: ValidationLevel


class WorkerClassResources(BaseModel):
    cores: int = Field(..., title="Cores")
    memory_mb: int = Field(..., title="Memory Mb")
    gpus: Optional[int] = Field(0, title="Gpus")
    gpu_memory_mb: Optional[int] = Field(None, title="Gpu Memory Mb")


class WorkerClassSpec(BaseModel):
    name: str = Field(..., title="Name")
    resources: WorkerClassResources
    description: str = Field(..., title="Description")
    instance_type: str = Field(..., title="Instance Type")
    cloud_provider: Optional[str] = Field("aws", title="Cloud Provider")


class ActionCreating(BaseModel):
    project_id: UUID = Field(..., title="Project Id")
    name: str = Field(..., title="Name")
    evaluation: str = Field(..., title="Evaluation")
    plan: RecipePlanCreating


class ActionExecution(BaseModel):
    id: UUID = Field(..., title="Id")
    is_running: bool = Field(..., title="Is Running")
    started: datetime = Field(..., title="Started")
    runtime: Optional[float] = Field(None, title="Runtime")
    state: ActionState


class ActionReading(BaseModel):
    id: Optional[UUID] = Field(None, title="Id")
    name: Optional[str] = Field(None, title="Name")
    state: Optional[ActionState] = None
    project_id: Optional[UUID] = Field(None, title="Project Id")
    broker_id: Optional[UUID] = Field(None, title="Broker Id")


class ActionSummary(BaseModel):
    id: UUID = Field(..., title="Id")
    created: datetime = Field(..., title="Created")
    updated: datetime = Field(..., title="Updated")
    is_running: bool = Field(..., title="Is Running")
    is_startable: bool = Field(..., title="Is Startable")
    is_stoppable: bool = Field(..., title="Is Stoppable")
    broker_id: UUID = Field(..., title="Broker Id")
    name: str = Field(..., title="Name")
    project_id: UUID = Field(..., title="Project Id")
    job_type: JobType
    state: ActionState
    plan: RecipePlanSummary
    cli_command: str = Field(..., title="Cli Command")
    stats: ActionStats


class AssetDetail(BaseModel):
    id: UUID = Field(..., title="Id")
    created: datetime = Field(..., title="Created")
    updated: datetime = Field(..., title="Updated")
    broker_id: UUID = Field(..., title="Broker Id")
    name: str = Field(..., title="Name")
    version: str = Field(..., title="Version")
    kind: str = Field(..., title="Kind")
    path: str = Field(..., title="Path")
    meta: Dict[str, Any] = Field(..., title="Meta")
    num_used_by: int = Field(..., title="Num Used By")
    tasks: List[TaskSummary] = Field(..., title="Tasks")
    actions: List[ActionSummary] = Field(..., title="Actions")


class BrokerDetail(BaseModel):
    id: UUID = Field(..., title="Id")
    created: datetime = Field(..., title="Created")
    updated: datetime = Field(..., title="Updated")
    org_id: UUID = Field(..., title="Org Id")
    name: str = Field(..., title="Name")
    address: str = Field(..., title="Address")
    state: str = Field(..., title="State")
    cloud_provider: str = Field(..., title="Cloud Provider")
    cloud_account: Optional[str] = Field(None, title="Cloud Account")
    cloud_region: Optional[str] = Field(None, title="Cloud Region")
    client_id: UUID = Field(..., title="Client Id")
    client_secret: str = Field(..., title="Client Secret")
    worker_classes: List[WorkerClassSpec] = Field(..., title="Worker Classes")


class ClusterJobReading(BaseModel):
    job_id: UUID = Field(..., title="Job Id")
    user_id: UUID = Field(..., title="User Id")
    job_type: JobType


class ClusterJobReturning(BaseModel):
    job_id: UUID = Field(..., title="Job Id")
    execution_id: Optional[UUID] = Field(None, title="Execution Id")
    attempt_id: UUID = Field(..., title="Attempt Id")
    job_type: JobType
    name: str = Field(..., title="Name")
    plan: RecipePlanDetail
    pam_access_token: str = Field(..., title="Pam Access Token")
    constraints: Optional[List[Constraint]] = Field(None, title="Constraints")
    resources: Optional[Resources] = None
    prodigy_jwt_secret: Optional[str] = Field(None, title="Prodigy Jwt Secret")


class DatasetDetail(BaseModel):
    id: UUID = Field(..., title="Id")
    created: datetime = Field(..., title="Created")
    updated: datetime = Field(..., title="Updated")
    name: str = Field(..., title="Name")
    broker_id: UUID = Field(..., title="Broker Id")
    kind: str = Field(..., title="Kind")
    num_used_by: int = Field(..., title="Num Used By")
    tasks: List[TaskSummary] = Field(..., title="Tasks")
    actions: List[ActionSummary] = Field(..., title="Actions")


class GroupDetail(BaseModel):
    id: UUID = Field(..., title="Id")
    created: datetime = Field(..., title="Created")
    updated: datetime = Field(..., title="Updated")
    name: str = Field(..., title="Name")
    org_id: UUID = Field(..., title="Org Id")
    users: List[UserDetail] = Field(..., title="Users")


class HTTPValidationError(BaseModel):
    detail: Optional[List[ValidationError]] = Field(None, title="Detail")


class NomadEventBatchCreating(BaseModel):
    index: int = Field(..., title="Index")
    broker_id: UUID = Field(..., title="Broker Id")
    events: List[NomadEventCreating] = Field(..., title="Events")


class NomadTaskEvent(BaseModel):
    Type: TaskEventType
    Time: int = Field(..., title="Time")
    Message: str = Field(..., title="Message")
    DisplayMessage: str = Field(..., title="Displaymessage")
    Details: TaskEventDetails
    FailsTask: bool = Field(..., title="Failstask")
    RestartReason: str = Field(..., title="Restartreason")
    SetupError: str = Field(..., title="Setuperror")
    DriverError: str = Field(..., title="Drivererror")
    ExitCode: int = Field(..., title="Exitcode")
    Signal: int = Field(..., title="Signal")
    KillTimeout: int = Field(..., title="Killtimeout")
    KillError: str = Field(..., title="Killerror")
    KillReason: str = Field(..., title="Killreason")
    StartDelay: int = Field(..., title="Startdelay")
    DownloadError: str = Field(..., title="Downloaderror")
    ValidationError: str = Field(..., title="Validationerror")
    DiskLimit: int = Field(..., title="Disklimit")
    FailedSibling: str = Field(..., title="Failedsibling")
    VaultError: str = Field(..., title="Vaulterror")
    TaskSignalReason: str = Field(..., title="Tasksignalreason")
    TaskSignal: str = Field(..., title="Tasksignal")
    DriverMessage: str = Field(..., title="Drivermessage")
    GenericSource: str = Field(..., title="Genericsource")


class NomadTaskState(BaseModel):
    State: str = Field(..., title="State")
    Failed: bool = Field(..., title="Failed")
    Restarts: int = Field(..., title="Restarts")
    LastRestart: Optional[int] = Field(None, title="Lastrestart")
    StartedAt: Optional[datetime] = Field(None, title="Startedat")
    FinishedAt: Optional[datetime] = Field(None, title="Finishedat")
    Events: List[NomadTaskEvent] = Field(..., title="Events")


class PackageDetail(BaseModel):
    id: UUID = Field(..., title="Id")
    created: datetime = Field(..., title="Created")
    updated: datetime = Field(..., title="Updated")
    name: str = Field(..., title="Name")
    version: str = Field(..., title="Version")
    environment: str = Field(..., title="Environment")
    num_used_by: int = Field(..., title="Num Used By")
    recipe_count: int = Field(..., title="Recipe Count")
    org_id: UUID = Field(..., title="Org Id")
    filename: str = Field(..., title="Filename")
    meta: Dict[str, Any] = Field(..., title="Meta")
    tasks: List[TaskSummary] = Field(..., title="Tasks")
    actions: List[ActionSummary] = Field(..., title="Actions")
    author: Optional[str] = Field(None, title="Author")


class PageActionSummary(BaseModel):
    items: List[ActionSummary] = Field(..., title="Items")
    total: int = Field(..., ge=0, title="Total")
    page: int = Field(..., ge=1, title="Page")
    size: int = Field(..., ge=1, title="Size")


class PagePersonSummary(BaseModel):
    items: List[PersonSummary] = Field(..., title="Items")
    total: int = Field(..., ge=0, title="Total")
    page: int = Field(..., ge=1, title="Page")
    size: int = Field(..., ge=1, title="Size")


class PageProjectSummary(BaseModel):
    items: List[ProjectSummary] = Field(..., title="Items")
    total: int = Field(..., ge=0, title="Total")
    page: int = Field(..., ge=1, title="Page")
    size: int = Field(..., ge=1, title="Size")


class PageRecipePlanSummary(BaseModel):
    items: List[RecipePlanSummary] = Field(..., title="Items")
    total: int = Field(..., ge=0, title="Total")
    page: int = Field(..., ge=1, title="Page")
    size: int = Field(..., ge=1, title="Size")


class PageRecipeSummary(BaseModel):
    items: List[RecipeSummary] = Field(..., title="Items")
    total: int = Field(..., ge=0, title="Total")
    page: int = Field(..., ge=1, title="Page")
    size: int = Field(..., ge=1, title="Size")


class PageSecretSummary(BaseModel):
    items: List[SecretSummary] = Field(..., title="Items")
    total: int = Field(..., ge=0, title="Total")
    page: int = Field(..., ge=1, title="Page")
    size: int = Field(..., ge=1, title="Size")


class PageSessionSummary(BaseModel):
    items: List[SessionSummary] = Field(..., title="Items")
    total: int = Field(..., ge=0, title="Total")
    page: int = Field(..., ge=1, title="Page")
    size: int = Field(..., ge=1, title="Size")


class PageTaskAnnotator(BaseModel):
    items: List[TaskAnnotator] = Field(..., title="Items")
    total: int = Field(..., ge=0, title="Total")
    page: int = Field(..., ge=1, title="Page")
    size: int = Field(..., ge=1, title="Size")


class PageTaskSummary(BaseModel):
    items: List[TaskSummary] = Field(..., title="Items")
    total: int = Field(..., ge=0, title="Total")
    page: int = Field(..., ge=1, title="Page")
    size: int = Field(..., ge=1, title="Size")


class PageUserSummary(BaseModel):
    items: List[UserSummary] = Field(..., title="Items")
    total: int = Field(..., ge=0, title="Total")
    page: int = Field(..., ge=1, title="Page")
    size: int = Field(..., ge=1, title="Size")


class Props(BaseModel):
    name: Optional[str] = Field(None, title="Name")
    title: Optional[str] = Field(None, title="Title")
    description: Optional[str] = Field(None, title="Description")
    placeholder: Optional[str] = Field(None, title="Placeholder")
    value: Optional[Any] = Field(None, title="Value")
    required: Optional[bool] = Field(None, title="Required")
    positional: Optional[bool] = Field(None, title="Positional")
    exists: Optional[bool] = Field(None, title="Exists")
    validations: Optional[List[ValidationSchema]] = Field(None, title="Validations")
    pattern: Optional[str] = Field(None, title="Pattern")
    widget: Optional[str] = Field(None, title="Widget")
    choice: Optional[Dict[str, Props]] = Field(None, title="Choice")
    min: Optional[float] = Field(None, title="Min")
    max: Optional[float] = Field(None, title="Max")
    step: Optional[float] = Field(None, title="Step")
    optional_title: Optional[str] = Field(None, title="Optional Title")


class TaskDetail(BaseModel):
    id: UUID = Field(..., title="Id")
    created: datetime = Field(..., title="Created")
    updated: datetime = Field(..., title="Updated")
    project_id: UUID = Field(..., title="Project Id")
    broker_id: UUID = Field(..., title="Broker Id")
    name: str = Field(..., title="Name")
    recipe_name: str = Field(..., title="Recipe Name")
    recipe_title: str = Field(..., title="Recipe Title")
    state: TaskState
    project_name: str = Field(..., title="Project Name")
    is_running: bool = Field(..., title="Is Running")
    is_startable: bool = Field(..., title="Is Startable")
    is_stoppable: bool = Field(..., title="Is Stoppable")
    error: Optional[str] = Field(None, title="Error")
    plan: RecipePlanDetail
    job_type: JobType
    cli_command: str = Field(..., title="Cli Command")
    nomad_index: int = Field(..., title="Nomad Index")
    last_execution_id: Optional[UUID] = Field(None, title="Last Execution Id")
    related_actions: List[UUID] = Field(..., title="Related Actions")
    url_logs: str = Field(..., title="Url Logs")
    url: str = Field(..., title="Url")
    created_by_user: Optional[UserSummary] = None


class TaskReading(BaseModel):
    id: Optional[UUID] = Field(None, title="Id")
    name: Optional[str] = Field(None, title="Name")
    state: Optional[TaskState] = None
    project_id: Optional[UUID] = Field(None, title="Project Id")


class ActionDetail(BaseModel):
    id: UUID = Field(..., title="Id")
    created: datetime = Field(..., title="Created")
    updated: datetime = Field(..., title="Updated")
    is_running: bool = Field(..., title="Is Running")
    is_startable: bool = Field(..., title="Is Startable")
    is_stoppable: bool = Field(..., title="Is Stoppable")
    broker_id: UUID = Field(..., title="Broker Id")
    name: str = Field(..., title="Name")
    project_id: UUID = Field(..., title="Project Id")
    job_type: JobType
    state: ActionState
    plan: RecipePlanDetail
    cli_command: str = Field(..., title="Cli Command")
    stats: ActionStats
    recipe_name: str = Field(..., title="Recipe Name")
    recipe_title: str = Field(..., title="Recipe Title")
    evaluation: str = Field(..., title="Evaluation")
    project_name: str = Field(..., title="Project Name")
    error: Optional[str] = Field(None, title="Error")
    executions: List[ActionExecution] = Field(..., title="Executions")
    last_execution_id: Optional[UUID] = Field(None, title="Last Execution Id")
    nomad_index: int = Field(..., title="Nomad Index")
    url_logs: str = Field(..., title="Url Logs")
    url: str = Field(..., title="Url")
    related_tasks: List[UUID] = Field(..., title="Related Tasks")
    created_by_user: Optional[UserSummary] = None


class FieldSchema(BaseModel):
    name: str = Field(..., title="Name")
    type: TypeSchema
    props: Props
    fields: Optional[List[FieldSchema]] = Field([], title="Fields")
    choice: Optional[List[FieldSchema]] = Field([], title="Choice")


class NomadAllocationStub(BaseModel):
    ID: UUID = Field(..., title="Id")
    EvalID: UUID = Field(..., title="Evalid")
    Name: str = Field(..., title="Name")
    Namespace: str = Field(..., title="Namespace")
    NodeID: UUID = Field(..., title="Nodeid")
    NodeName: str = Field(..., title="Nodename")
    JobID: UUID = Field(..., title="Jobid")
    JobType: Optional[str] = Field(None, title="Jobtype")
    JobVersion: Optional[int] = Field(None, title="Jobversion")
    TaskGroup: str = Field(..., title="Taskgroup")
    DesiredStatus: AllocDesiredStatus
    DesiredDescription: Optional[str] = Field(None, title="Desireddescription")
    ClientStatus: AllocClientStatus
    ClientDescription: Optional[str] = Field(None, title="Clientdescription")
    TaskStates: Optional[Dict[str, NomadTaskState]] = Field({}, title="Taskstates")
    FollowupEvalID: Optional[UUID] = Field(None, title="Followupevalid")
    PreemptedAllocations: Optional[List[str]] = Field(
        None, title="Preemptedallocations"
    )
    PreemptedByAllocation: Optional[UUID] = Field(None, title="Preemptedbyallocation")
    CreateIndex: int = Field(..., title="Createindex")
    ModifyIndex: int = Field(..., title="Modifyindex")
    CreateTime: int = Field(..., title="Createtime")
    ModifyTime: int = Field(..., title="Modifytime")


class RecipeSchema(BaseModel):
    name: str = Field(..., title="Name")
    fields: List[FieldSchema] = Field(..., title="Fields")
    title: Optional[str] = Field(None, title="Title")
    description: Optional[str] = Field(None, title="Description")
    view_id: Optional[str] = Field(None, title="View Id")
    cli_names: Optional[Dict[str, str]] = Field(None, title="Cli Names")


class NomadJobs(BaseModel):
    jobs: List[NomadJobStub] = Field(..., title="Jobs")
    allocations: List[NomadAllocationStub] = Field(..., title="Allocations")


class RecipeCreating(BaseModel):
    name: str = Field(..., title="Name")
    title: str = Field(..., title="Title")
    description: str = Field(..., title="Description")
    is_action: bool = Field(..., title="Is Action")
    entry_point: str = Field(..., title="Entry Point")
    form_schema: RecipeSchema
    cli_schema: Dict[str, Any] = Field(..., title="Cli Schema")
    meta: Dict[str, Any] = Field(..., title="Meta")
    package_id: UUID = Field(..., title="Package Id")


class RecipeDetail(BaseModel):
    id: UUID = Field(..., title="Id")
    created: datetime = Field(..., title="Created")
    updated: datetime = Field(..., title="Updated")
    org_id: UUID = Field(..., title="Org Id")
    package_id: UUID = Field(..., title="Package Id")
    name: str = Field(..., title="Name")
    title: Optional[str] = Field(None, title="Title")
    description: Optional[str] = Field(None, title="Description")
    is_action: bool = Field(..., title="Is Action")
    entry_point: Optional[str] = Field(None, title="Entry Point")
    num_used_by: int = Field(..., title="Num Used By")
    form_schema: RecipeSchema
    cli_schema: Dict[str, Any] = Field(..., title="Cli Schema")
    meta: Dict[str, Any] = Field(..., title="Meta")
    tasks: List[TaskSummary] = Field(..., title="Tasks")
    actions: List[ActionSummary] = Field(..., title="Actions")
    package: PackageSummary


CackQuerying.update_forward_refs()
TypeSchema.update_forward_refs()
Props.update_forward_refs()
FieldSchema.update_forward_refs()



from ..models import (
    NotificationCreating,
    NotificationDeleting,
    NotificationDetail,
    NotificationReading,
    NotificationSummary,
    NotificationUpdating,
)
from .base import ModelClient


class Notification(
    ModelClient[
        NotificationCreating,
        NotificationReading,
        NotificationUpdating,
        NotificationDeleting,
        NotificationSummary,
        NotificationDetail,
    ]
):
    Creating = NotificationCreating
    Reading = NotificationReading
    Updating = NotificationUpdating
    Deleting = NotificationDeleting
    Summary = NotificationSummary
    Detail = NotificationDetail



from .. import ty
from ..models import Secret, SecretResponse
from .base import BaseClient


class Consul(BaseClient):
    def create(self, secret: Secret) -> bool:
        res = self.request("POST", endpoint="create", data=secret)
        return ty.cast(bool, res)

    async def create_async(self, secret: Secret) -> bool:
        res = await self.request_async("POST", endpoint="create", data=secret)
        return ty.cast(bool, res)

    def read(self, key: str) -> ty.List[SecretResponse]:
        res = self.request("POST", endpoint="read", data={"key": key})
        return [SecretResponse.parse_obj(obj) for obj in ty.cast(ty.List[dict], res)]

    async def read_async(self, key: str) -> ty.List[SecretResponse]:
        res = await self.request_async("POST", endpoint="read", data={"key": key})
        return [SecretResponse.parse_obj(obj) for obj in ty.cast(ty.List[dict], res)]

    def delete(self, key: str) -> bool:
        res = self.request("POST", endpoint="delete", data={"key": key})
        return ty.cast(bool, res)

    async def delete_async(self, key: str) -> bool:
        res = await self.request_async("POST", endpoint="delete", data={"key": key})
        return ty.cast(bool, res)



from typing import cast

from ..models import (
    SessionCreating,
    SessionDeleting,
    SessionDetail,
    SessionReading,
    SessionSummary,
    SessionTokenReturning,
    SessionUpdating,
)
from .base import ModelClient


class Session(
    ModelClient[
        SessionCreating,
        SessionReading,
        SessionUpdating,
        SessionDeleting,
        SessionSummary,
        SessionDetail,
    ]
):
    Creating = SessionCreating
    Reading = SessionReading
    Updating = SessionUpdating
    Deleting = SessionDeleting
    Summary = SessionSummary
    Detail = SessionDetail

    def token(self, data: SessionCreating) -> SessionTokenReturning:
        res = self.request(
            "POST",
            endpoint="token",
            data=data,
            return_model=SessionTokenReturning,
        )
        return cast(SessionTokenReturning, res)

    async def token_async(self, data: SessionCreating) -> SessionTokenReturning:
        res = await self.request_async(
            "POST",
            endpoint="token",
            data=data,
            return_model=SessionTokenReturning,
        )
        return cast(SessionTokenReturning, res)



import os
import os.path
from itertools import chain
from pathlib import Path

from radicli import Arg

from ... import ty
from ...cli import cli
from ...errors import BrokerError, CLIError
from ...messages import Messages
from ...prodigy_teams_broker_sdk import Client as BrokerClient
from ...prodigy_teams_broker_sdk import models as broker_models
from ...ui import Progress
from ...util import _resolve_broker_ref, is_local, resolve_remote_path
from .._state import get_auth_state, get_saved_settings


class RsyncCallable(ty.Protocol):
    def __call__(
        self,
        *,
        src: str,
        dest: str,
        broker_client: BrokerClient,
    ) -> ty.Optional[str]:
        ...


@cli.subcommand(
    "files",
    "rsync",
    src=Arg(help=Messages.remote_local_path.format(noun="source")),
    dest=Arg(help=Messages.remote_local_path.format(noun="destination")),
    cluster_host=Arg("--cluster-host", help=Messages.cluster_host),
)
def rsync(src: str, dest: str, cluster_host: ty.Optional[str] = None) -> None:
    """Rsync files to and from the cluster"""
    settings = get_saved_settings()
    auth = get_auth_state()
    broker_host = str(
        _resolve_broker_ref(auth.client, cluster_host or settings.broker_host)
    )
    is_src_local = is_local(src)
    is_dest_local = is_local(dest)

    def _rsync(f: RsyncCallable, *, src, dest) -> ty.Optional[str]:
        return f(
            src=src,
            dest=dest,
            broker_client=auth.broker_client,
        )

    if is_src_local and is_dest_local:
        raise CLIError(Messages.E015.format(verb="rsync"), f"{src} -> {dest}")
    elif is_src_local and not is_dest_local:
        _rsync(
            _rsync_upload,
            src=src,
            dest=resolve_remote_path(auth.client, dest, broker_host),
        )
    elif not is_src_local and is_dest_local:
        _rsync(
            _rsync_download,
            src=resolve_remote_path(auth.client, src, broker_host),
            dest=dest,
        )
    else:
        _rsync(
            _rsync_on_remote,
            src=resolve_remote_path(auth.client, src, broker_host),
            dest=resolve_remote_path(auth.client, dest, broker_host),
        )


T1 = ty.TypeVar("T1")
T2 = ty.TypeVar("T2")
R = ty.TypeVar("R")


def diff(
    left: ty.List[T1],
    right: ty.List[T2],
    left_key: ty.Callable[[T1], R],
    right_key: ty.Callable[[T2], R],
) -> ty.Tuple[ty.List[T1], ty.List[T2]]:
    left_idx = {left_key(value): value for value in left}
    right_idx = {right_key(value): value for value in right}
    left_result = [left_idx[k] for k in left_idx.keys() - right_idx.keys()]
    right_result = [right_idx[k] for k in right_idx.keys() - left_idx.keys()]
    return left_result, right_result


def _rsync_upload(
    *,
    broker_client: BrokerClient,
    src: str,
    dest: str,
) -> None:
    src_path = Path(src).resolve().absolute()
    src = str(src_path)
    src_paths = sorted(
        [
            str(path.resolve().absolute())
            for path in chain((src_path,), src_path.glob("**/*"))
            if path.is_file()
        ],
        key=lambda path: path.removeprefix(src),
    )
    body = broker_models.Listing(path=dest, recurse=True, include_stats=False)
    try:
        files = broker_client.files.list_dir(body)
    except BrokerError as e:
        raise CLIError(Messages.E018, e)
    dest_paths = sorted(
        files.paths,
        key=lambda path: path.removeprefix(dest),
    )
    to_copy, to_delete = diff(
        src_paths,
        dest_paths,
        left_key=lambda path: path.removeprefix(src),
        right_key=lambda path: path.removeprefix(dest),
    )
    _rsync_upload_util(
        src_base=src,
        dest_base=dest,
        to_copy=to_copy,
        to_delete=to_delete,
        broker_client=broker_client,
    )


def _rsync_upload_util(
    *,
    src_base: str,
    dest_base: str,
    to_copy: ty.List[str],
    to_delete: ty.List[str],
    broker_client,
) -> None:
    def get_dest(path: str) -> str:
        return os.path.join(dest_base, path.removeprefix(src_base).removeprefix("/"))

    with Progress(to_copy) as paths:
        for path in paths:
            body = Path(path).open("rb").read()
            try:
                broker_client.files.upload(
                    body, dest=get_dest(path), overwrite=True, make_dirs=True
                )
            except BrokerError as e:
                raise CLIError(Messages.E017.format(verb="upload"), e)
    with Progress(to_delete) as paths:
        for path in paths:
            body = broker_models.Deleting(path=path, missing_ok=False, recurse=False)
            try:
                broker_client.files.delete(body)
            except BrokerError as e:
                raise CLIError(Messages.E017.format(verb="delete"), e)


def _rsync_download(
    *,
    broker_client: BrokerClient,
    src: str,
    dest: str,
) -> None:
    body = broker_models.Listing(path=src, recurse=True, include_stats=False)
    try:
        paths = broker_client.files.list_dir(body)
    except BrokerError as e:
        raise CLIError(Messages.E018, e)
    src_paths = sorted(paths.paths)
    if src_paths == src:
        raise CLIError(Messages.E024, src)
    dest_path = Path(dest).resolve().absolute()
    dest = str(dest_path)
    dest_paths = sorted(
        [
            str(path.resolve().absolute())
            for path in chain((dest_path,), dest_path.glob("**/*"))
            if path.is_file()
        ],
    )
    to_copy, to_delete = diff(
        src_paths,
        dest_paths,
        left_key=lambda path: path.removeprefix(src),
        right_key=lambda path: path.removeprefix(dest),
    )
    _rsync_download_util(
        src_base=src,
        dest_base=dest,
        to_copy=to_copy,
        to_delete=to_delete,
        broker_client=broker_client,
    )


def _rsync_download_util(
    *,
    src_base: str,
    dest_base: str,
    to_copy: ty.List[str],
    to_delete: ty.List[str],
    broker_client: BrokerClient,
) -> None:
    def get_dest(path: str) -> str:
        return os.path.join(dest_base, path.removeprefix(src_base).removeprefix("/"))

    with Progress(to_copy) as paths:
        for path in paths:
            body = broker_models.Downloading(target=path)
            try:
                content = broker_client.files.download(body)
            except BrokerError as e:
                raise CLIError(Messages.E017.format(verb="download"), e)
            dest = Path(get_dest(path))
            dest.parent.mkdir(parents=True, exist_ok=True)
            with dest.open("wb") as file_:
                file_.write(content.read())
    with Progress(to_delete) as paths:
        for path in paths:
            path = Path(path)
            if path.exists():
                path.unlink()


def _rsync_on_remote(
    *,
    broker_client: BrokerClient,
    src: str,
    dest: str,
) -> None:
    body = broker_models.Copying(src=src, dest=dest, make_dirs=True, overwrite=True)
    try:
        plan = broker_client.files.plan_directory_rsync(body)
    except BrokerError as e:
        raise CLIError(Messages.E022, e)
    with Progress(plan.copy_) as copies:
        for file_copy_plan in copies:
            body = broker_models.Copying(
                src=file_copy_plan.src,
                dest=file_copy_plan.dest,
                make_dirs=True,
                overwrite=True,
            )
            try:
                broker_client.files.copy(body)
            except BrokerError as e:
                raise CLIError(Messages.E017.format(verb="copy"), e)
    with Progress(plan.delete) as paths:
        for path in paths:
            body = broker_models.Deleting(path=path, missing_ok=False, recurse=False)
            try:
                broker_client.files.delete(body)
            except BrokerError as e:
                raise CLIError(Messages.E017.format(verb="delete"), e)



from datetime import datetime, timedelta

import httpx
import jwt

from .. import ty
from .base import BaseClient
from .check import Check
from .consul import Consul
from .data import Data
from .envs2 import Envs2
from .files import Files
from .jobs import Jobs


class ClientCredentialsAuthRequest(ty.NamedTuple):
    token_url: str
    auth: httpx.BasicAuth
    data: ty.Dict[str, str]


class AccessTokenCredential(ty.BaseModel):
    access_token: str

    @property
    def header(self) -> ty.Dict[str, ty.Any]:
        """
        Return the JWT header dict of the `self.access_token`.
        The signature is not verified, but this allows inspection of e.g. expiration
        """
        return jwt.decode(
            self.access_token,
            algorithms=["HS256"],
            options={"verify_signature": False},
        )

    @property
    def expires_at(self) -> datetime:
        return datetime.utcfromtimestamp(self.header["exp"])

    def expires_within(self, delta: timedelta) -> bool:
        """Return True if the access token expires within `seconds` seconds."""
        return datetime.utcnow() + delta > self.expires_at

    @property
    def expired(self) -> bool:
        """Returns True if the access token has expired."""
        return self.expires_within(timedelta(seconds=5))


class Client:
    check: Check
    consul: Consul
    data: Data
    envs2: Envs2
    files: Files
    jobs: Jobs

    def __init__(
        self,
        base_url: str,
        token: AccessTokenCredential,
        client_id: ty.Optional[str] = None,
        client_secret: ty.Optional[str] = None,
    ) -> None:
        self._base_url = base_url
        self._token = token
        self._client_id = client_id
        self._client_secret = client_secret
        self._auth_headers = {"Authorization": f"Bearer {token.access_token}"}
        self._sync_client = httpx.Client(base_url=base_url, headers=self._auth_headers)
        self._async_client = httpx.AsyncClient(
            base_url=base_url, headers=self._auth_headers
        )
        self._base_client = BaseClient(self._sync_client, self._async_client, "")
        self.request = self._base_client.request
        self.request_async = self._base_client.request_async
        self.check = Check(self._sync_client, self._async_client, "/api/v1/check")
        self.consul = Consul(self._sync_client, self._async_client, "/api/v1/consul")
        self.datasets = Data(self._sync_client, self._async_client, "/api/v1/datasets")
        self.envs2 = Envs2(self._sync_client, self._async_client, "/api/v1/envs2")
        self.files = Files(self._sync_client, self._async_client, "/api/v1/files")
        self.jobs = Jobs(self._sync_client, self._async_client, "/api/v1/jobs")

    def close(self) -> None:
        self._sync_client.close()

    async def aclose(self) -> None:
        self._sync_client.close()
        await self._async_client.aclose()

    @property
    def base_url(self) -> str:
        return self._base_url

    @property
    def access_token(self) -> str:
        return self._token.access_token

    @classmethod
    def from_client_creds(
        cls, base_url: str, client_id: str, client_secret: str
    ) -> "Client":
        cred = cls._make_authorization_request(base_url, client_id, client_secret)
        return cls(
            base_url=base_url,
            token=cred,
            client_id=client_id,
            client_secret=client_secret,
        )

    @classmethod
    async def from_client_creds_async(
        cls, base_url: str, client_id: str, client_secret: str
    ) -> "Client":
        cred = await cls._make_authorization_request_async(
            base_url, client_id, client_secret
        )
        return cls(
            base_url=base_url,
            token=cred,
            client_id=client_id,
            client_secret=client_secret,
        )

    def ensure_auth(self) -> None:
        if self._token.expired and self._client_id and self._client_secret:
            cred = self._make_authorization_request(
                self._base_url, self._client_id, self._client_secret
            )
            self._update_auth(cred)

    async def ensure_auth_async(self) -> None:
        if self._token.expired and self._client_id and self._client_secret:
            cred = await self._make_authorization_request_async(
                self._base_url, self._client_id, self._client_secret
            )
            self._update_auth(cred)

    def _update_auth(self, token: AccessTokenCredential) -> None:
        self._token = token
        self._auth_headers = {"Authorization": f"Bearer {token.access_token}"}
        self._async_client.headers.update(self._auth_headers)
        self._sync_client.headers.update(self._auth_headers)

    @staticmethod
    def _make_authorization_request(
        base_url: str, client_id: str, client_secret: str
    ) -> AccessTokenCredential:
        """Send a Client Credentials Auth request to Prodigy Teams to
        receive an authorized Access Token."""
        req = Client._create_auth_req(base_url, client_id, client_secret)
        res = httpx.post(req.token_url, auth=req.auth, data=req.data)
        return Client._handle_auth_res(res)

    @staticmethod
    async def _make_authorization_request_async(
        base_url: str, client_id: str, client_secret: str
    ) -> AccessTokenCredential:
        """Send a Client Credentials Auth request to Prodigy Teams to
        receive an authorized Access Token."""
        req = Client._create_auth_req(base_url, client_id, client_secret)
        async with httpx.AsyncClient(auth=req.auth) as http_client:
            res = await http_client.post(req.token_url, data=req.data)
        return Client._handle_auth_res(res)

    @staticmethod
    def _create_auth_req(
        base_url: str, client_id: str, client_secret: str
    ) -> ClientCredentialsAuthRequest:
        token_url = f"{base_url}/api/v1/oauth/token"
        auth = httpx.BasicAuth(username=client_id, password=client_secret)
        data = {"grant_type": "client_credentials"}
        return ClientCredentialsAuthRequest(token_url=token_url, auth=auth, data=data)

    @staticmethod
    def _handle_auth_res(res: httpx.Response) -> AccessTokenCredential:
        """Send a Client Credentials Auth request to Prodigy Teams to
        receive an authorized Access Token."""
        res.raise_for_status()
        data = res.json()
        return AccessTokenCredential(access_token=data["access_token"])



from ..models import (
    TaskCreating,
    TaskDeleting,
    TaskDetail,
    TaskReading,
    TaskSummary,
    TaskUpdating,
)
from .base import ModelClient


class Task(
    ModelClient[
        TaskCreating,
        TaskReading,
        TaskUpdating,
        TaskDeleting,
        TaskSummary,
        TaskDetail,
    ]
):
    Creating = TaskCreating
    Reading = TaskReading
    Updating = TaskUpdating
    Deleting = TaskDeleting
    Summary = TaskSummary
    Detail = TaskDetail



from ..models import (
    ProjectCreating,
    ProjectDeleting,
    ProjectDetail,
    ProjectReading,
    ProjectSummary,
    ProjectUpdating,
)
from .base import ModelClient


class Project(
    ModelClient[
        ProjectCreating,
        ProjectReading,
        ProjectUpdating,
        ProjectDeleting,
        ProjectSummary,
        ProjectDetail,
    ]
):
    Creating = ProjectCreating
    Reading = ProjectReading
    Updating = ProjectUpdating
    Deleting = ProjectDeleting
    Summary = ProjectSummary
    Detail = ProjectDetail



import httpx

from .. import ty
from ..models import EnvAwaiting, EnvCreating, EnvReturning
from .base import BaseClient


class Envs2(BaseClient):
    def create(self, body: EnvCreating) -> EnvReturning:
        res = self.request(
            "POST",
            endpoint="create",
            data=body,
            return_model=EnvReturning,
            timeout=httpx.Timeout(5.0, read=30.0),
        )
        return ty.cast(EnvReturning, res)

    async def create_async(self, body: EnvCreating) -> EnvReturning:
        res = self.request_async(
            "POST",
            endpoint="create",
            data=body,
            return_model=EnvReturning,
            timeout=httpx.Timeout(5.0, read=30.0),
        )
        return ty.cast(EnvReturning, res)

    def await_env(self, body: EnvAwaiting) -> EnvReturning:
        res = self.request(
            "POST",
            endpoint="await",
            data=body,
            return_model=EnvReturning,
            timeout=httpx.Timeout(5.0, read=30.0),
        )
        return ty.cast(EnvReturning, res)

    async def await_env_async(self, body: EnvAwaiting) -> EnvReturning:
        res = self.request_async(
            "POST",
            endpoint="await",
            data=body,
            return_model=EnvReturning,
            timeout=httpx.Timeout(5.0, read=30.0),
        )
        return ty.cast(EnvReturning, res)



import json

import httpx
from pydantic import BaseModel

from .. import errors, ty


class BaseClient:
    name: str
    url: str

    def __init__(
        self, sync_client: httpx.Client, async_client: httpx.AsyncClient, path: str
    ) -> None:
        self._sync_client = sync_client
        self._async_client = async_client
        if path.startswith("/"):
            path = path[1:]
        self.name = path.split("/")[-1]
        self.path = f"/{path}"

    def request(
        self,
        method: str,
        *,
        endpoint: ty.Optional[str] = None,
        params: ty.Union[None, BaseModel, ty.Dict[str, ty.Any]] = None,
        data: ty.Union[None, BaseModel, ty.Dict[str, ty.Any]] = None,
        headers: ty.Dict[str, str] = {},
        page: ty.Optional[int] = None,
        size: ty.Optional[int] = None,
        params_model: ty.Optional[ty.Type[BaseModel]] = None,
        body_model: ty.Optional[ty.Type[BaseModel]] = None,
        return_model: ty.Optional[ty.Type[BaseModel]] = None,
    ) -> ty.Union[None, BaseModel, ty.Page[BaseModel]]:
        req = self._get_validated_request(
            method=method,
            endpoint=endpoint or "",
            params=params,
            data=data,
            params_model=params_model,
            body_model=body_model,
            headers=headers,
            page=page,
            size=size,
        )
        response = self._sync_client.send(req)
        self._raise_and_handle_errors(response)
        content = None
        if response.content:
            content = response.json()
        if return_model is not None:
            return self._convert_response_to_model(content, return_model)

    async def request_async(
        self,
        method: str,
        *,
        endpoint: ty.Optional[str] = None,
        params: ty.Union[None, BaseModel, ty.Dict[str, ty.Any]] = None,
        data: ty.Union[None, BaseModel, ty.Dict[str, ty.Any]] = None,
        headers: ty.Dict[str, str] = {},
        page: ty.Optional[int] = None,
        size: ty.Optional[int] = None,
        params_model: ty.Optional[ty.Type[BaseModel]] = None,
        body_model: ty.Optional[ty.Type[BaseModel]] = None,
        return_model: ty.Optional[ty.Type[BaseModel]] = None,
    ) -> ty.Union[None, BaseModel, ty.Page[BaseModel]]:
        req = self._get_validated_request(
            method=method,
            endpoint=endpoint or "",
            params=params,
            data=data,
            params_model=params_model,
            body_model=body_model,
            headers=headers,
            page=page,
            size=size,
        )
        response = await self._async_client.send(req)
        self._raise_and_handle_errors(response)
        content = None
        if response.content:
            content = response.json()
        if return_model is not None:
            return self._convert_response_to_model(content, return_model)

    def _get_validated_request(
        self,
        *,
        method: str,
        endpoint: str,
        params: ty.Union[None, BaseModel, ty.Dict[str, ty.Any]] = None,
        data: ty.Union[None, BaseModel, ty.Dict[str, ty.Any]] = None,
        params_model: ty.Optional[ty.Type[BaseModel]] = None,
        body_model: ty.Optional[ty.Type[BaseModel]] = None,
        headers: ty.Optional[ty.Dict[str, str]] = None,
        page: ty.Optional[int] = None,
        size: ty.Optional[int] = None,
    ) -> httpx.Request:
        url = self.path
        if endpoint:
            if endpoint.startswith("/"):
                url = endpoint
            else:
                url = f"{self.path}/{endpoint}"
        local_params = {}
        if page is not None:
            local_params["page"] = page
        if size is not None:
            local_params["size"] = size
        if params is not None:
            if isinstance(params, dict):
                params.update(local_params)
                if params_model is not None:
                    # if params_model is passed, use Pydantic to validate params
                    params = params_model(**params)
                    params = params.dict()
            elif isinstance(params, BaseModel):
                params = params.dict(exclude_none=True, exclude_defaults=True)
                params.update(local_params)
        content = None
        if data is not None:
            if isinstance(data, dict):
                if body_model is not None:
                    body = body_model(**data)
                    content = body.json()
                else:
                    content = json.dumps(data)
            elif isinstance(data, BaseModel):
                content = data.json()
        req = self._sync_client.build_request(
            method, url, headers=headers, params=params, content=content
        )
        return req

    def _raise_and_handle_errors(self, response: httpx.Response) -> None:
        try:
            response.raise_for_status()
        except httpx.HTTPStatusError as e:
            status = e.response.status_code
            error_type = None
            sdk_error = None
            try:
                error_data = e.response.json()
            except json.JSONDecodeError:
                error_data = {}
            if isinstance(error_data, dict):
                detail = error_data.get("detail")
            else:
                detail = error_data
            if detail and isinstance(detail, dict):
                error_type = detail.get("type")
            if error_type:
                sdk_error = getattr(errors, error_type, None)
            if sdk_error:
                raise sdk_error(detail=str(detail))
            elif status == 401:
                raise errors.AuthError(detail=str(detail))
            else:
                raise e

    def _convert_response_to_model(
        self, data: ty.Any, return_model: ty.Type[BaseModel]
    ) -> ty.Union[None, BaseModel, ty.Page[BaseModel]]:
        """Convert the JSON response to the correct Pydantic model"""
        if data is None:
            return data
        return return_model.parse_obj(data)


_Creating = ty.TypeVar("_Creating", bound=BaseModel)
_Reading = ty.TypeVar("_Reading", bound=BaseModel)
_Updating = ty.TypeVar("_Updating", bound=BaseModel)
_Deleting = ty.TypeVar("_Deleting", bound=BaseModel)
_Summary = ty.TypeVar("_Summary", bound=BaseModel)
_Detail = ty.TypeVar("_Detail", bound=BaseModel)


class ModelClient(
    BaseClient,
    ty.Generic[_Creating, _Reading, _Updating, _Deleting, _Summary, _Detail],
):
    Creating: ty.Type[_Creating]
    Reading: ty.Type[_Reading]
    Updating: ty.Type[_Updating]
    Deleting: ty.Type[_Deleting]
    Summary: ty.Type[_Summary]
    Detail: ty.Type[_Detail]

    def __init__(
        self, sync_client: httpx.Client, async_client: httpx.AsyncClient, path: str
    ) -> None:
        super().__init__(sync_client, async_client, path)
        self.plural_path = f"{self.path}s"

    # TODO: Eventually the `create`, `read`, `update` routes should all
    # have _Detail models and they won't return Union[_Summary], _Detail] but
    # explicitly return `_Detail`. This is temporary until the PAM API is finalized
    def create(
        self, data: ty.Optional[_Creating] = None, **kwargs: ty.Any
    ) -> ty.Union[_Summary, _Detail]:
        data = data if data else self.Creating(**kwargs)
        has_detail, return_model = self._has_detail()
        # Using a type ignore here because the 'detail' stuff is messy and needs to be
        # refactored anyway
        res = self.request(
            "POST", endpoint="create", data=data, return_model=return_model  # type: ignore
        )
        return ty.cast(_Detail, res) if has_detail else ty.cast(_Summary, res)

    async def create_async(
        self, data: ty.Optional[_Creating] = None, **kwargs: ty.Any
    ) -> ty.Union[_Summary, _Detail]:
        data = data if data else self.Creating(**kwargs)
        has_detail, return_model = self._has_detail()
        res = await self.request_async(
            "POST", endpoint="create", data=data, return_model=return_model  # type: ignore
        )
        return ty.cast(_Detail, res) if has_detail else ty.cast(_Summary, res)

    def read(
        self, query: ty.Optional[_Reading] = None, **kwargs: ty.Any
    ) -> ty.Union[_Summary, _Detail]:
        params = query if query else self.Reading(**kwargs)
        has_detail, return_model = self._has_detail()
        res = self.request("GET", params=params, return_model=return_model)  # type: ignore
        return ty.cast(_Detail, res) if has_detail else ty.cast(_Summary, res)

    async def read_async(
        self, query: ty.Optional[_Reading] = None, **kwargs: ty.Any
    ) -> ty.Union[_Summary, _Detail]:
        params = query if query else self.Reading(**kwargs)
        has_detail, return_model = self._has_detail()
        res = await self.request_async("GET", params=params, return_model=return_model)  # type: ignore
        return ty.cast(_Detail, res) if has_detail else ty.cast(_Summary, res)

    def exists(
        self, query: ty.Optional[_Reading] = None, **kwargs: ty.Any
    ) -> ty.Optional[ty.Union[_Summary, _Detail]]:
        params = query if query else self.Reading(**kwargs)
        has_detail, return_model = self._has_detail()
        try:
            res = self.request("GET", params=params, return_model=return_model)  # type: ignore
        except errors.NotFound:
            res = None
        return (
            ty.cast(ty.Optional[_Detail], res)
            if has_detail
            else ty.cast(ty.Optional[_Summary], res)
        )

    async def exists_async(
        self, query: ty.Optional[_Reading] = None, **kwargs: ty.Any
    ) -> ty.Optional[ty.Union[_Summary, _Detail]]:
        params = query if query else self.Reading(**kwargs)
        has_detail, return_model = self._has_detail()
        try:
            res = await self.request_async(
                "GET", params=params, return_model=return_model  # type: ignore
            )
        except errors.NotFound:
            res = None
        return (
            ty.cast(ty.Optional[_Detail], res)
            if has_detail
            else ty.cast(ty.Optional[_Summary], res)
        )

    def update(self, data: ty.Optional[_Updating] = None, **kwargs: ty.Any) -> _Detail:
        data = data if data else self.Updating(**kwargs)
        has_detail, return_model = self._has_detail()
        res = self.request(
            "POST",
            endpoint="update",
            data=data,
            return_model=return_model,  # type: ignore
        )
        return res  # type: ignore

    async def update_async(
        self, data: ty.Optional[_Updating] = None, **kwargs: ty.Any
    ) -> _Summary:
        data = data if data else self.Updating(**kwargs)
        has_detail, return_model = self._has_detail()
        res = await self.request_async(
            "POST",
            endpoint="update",
            data=data,
            return_model=return_model,  # type: ignore
        )
        return res  # type: ignore

    def delete(self, **kwargs: ty.Any) -> None:
        self.request("POST", endpoint="delete", data=kwargs, body_model=self.Deleting)

    async def delete_async(self, **kwargs: ty.Any) -> None:
        await self.request_async(
            "POST", endpoint="delete", data=kwargs, body_model=self.Deleting
        )

    def all(
        self,
        query: ty.Optional[_Reading] = None,
        *,
        page: ty.Optional[int] = None,
        size: ty.Optional[int] = None,
        **kwargs: ty.Any,
    ) -> ty.Page[_Summary]:
        params = query if query else self.Reading(**kwargs)
        res = self.request(
            "GET",
            endpoint=self.plural_path,
            params=params,
            page=page,
            size=size,
            return_model=ty.Page[self.Summary],  # type: ignore
        )
        return ty.cast(ty.Page[_Summary], res)

    async def all_async(
        self,
        query: ty.Optional[_Reading] = None,
        *,
        page: ty.Optional[int] = None,
        size: ty.Optional[int] = None,
        **kwargs: ty.Any,
    ) -> ty.Page[_Summary]:
        params = query if query else self.Reading(**kwargs)
        res = await self.request_async(
            "GET",
            endpoint=self.plural_path,
            params=params,
            page=page,
            size=size,
            return_model=ty.Page[self.Summary],  # type: ignore
        )
        return ty.cast(ty.Page[_Summary], res)

    def all_readable(
        self,
        query: ty.Optional[_Reading] = None,
        **kwargs: ty.Any,
    ) -> ty.Page[_Summary]:
        params = query if query else self.Reading(**kwargs)
        res = self.request(
            "GET",
            endpoint=self.plural_path,
            params=params,
            return_model=ty.Page[self.Summary],  # type: ignore
        )
        return ty.cast(ty.Page[_Summary], res)

    async def all_readable_async(
        self,
        query: ty.Optional[_Reading] = None,
        **kwargs: ty.Any,
    ) -> ty.Page[_Summary]:
        params = query if query else self.Reading(**kwargs)
        res = await self.request_async(
            "GET",
            endpoint=self.plural_path,
            params=params,
            return_model=ty.Page[self.Summary],  # type: ignore
        )
        return ty.cast(ty.Page[_Summary], res)

    def _raise_and_handle_errors(self, response: httpx.Response) -> None:
        """Translate errors by status. This allows SDK consumers to get the original
        error from Prodigy Teams, not just the status.
        Does nothing if the reponse was successful
        """
        prefix = self.__class__.__name__
        try:
            super()._raise_and_handle_errors(response)
        except httpx.HTTPStatusError as e:
            status = e.response.status_code
            error_type = None
            sdk_error = None
            try:
                error_data = e.response.json()
            except json.JSONDecodeError:
                error_data = {}
            detail = error_data.get("detail")
            if detail and isinstance(detail, dict):
                error_type = detail.get("type")
            if error_type:
                sdk_error = getattr(errors, error_type, None)
            if sdk_error:
                raise sdk_error(detail=str(detail))
            elif status == 400:
                raise getattr(errors, f"{prefix}Invalid", errors.Invalid)
            elif status == 403:
                raise getattr(errors, f"{prefix}Forbidden", errors.Forbidden)
            elif status == 404:
                raise getattr(errors, f"{prefix}NotFound", errors.NotFound)
            else:
                raise e

    def _has_detail(self) -> ty.Tuple[bool, ty.Union[_Summary, _Detail]]:
        if hasattr(self, "Detail"):
            return True, ty.cast(_Detail, self.Detail)
        else:
            return False, ty.cast(_Detail, self.Summary)


_ReadingLatest = ty.TypeVar("_ReadingLatest", bound=BaseModel)
_ListingLatest = ty.TypeVar("_ListingLatest", bound=BaseModel)


class VersionedModelClient(
    ty.Generic[
        _Creating,
        _Reading,
        _Updating,
        _Deleting,
        _Summary,
        _Detail,
        _ReadingLatest,
        _ListingLatest,
    ],
    ModelClient[
        _Creating,
        _Reading,
        _Updating,
        _Deleting,
        _Summary,
        _Detail,
    ],
):
    Creating: ty.Type[_Creating]
    Reading: ty.Type[_Reading]
    Updating: ty.Type[_Updating]
    Deleting: ty.Type[_Deleting]
    Summary: ty.Type[_Summary]
    Detail: ty.Type[_Detail]
    ReadingLatest: ty.Type[_ReadingLatest]
    ListingLatest: ty.Type[_ListingLatest]

    def all_latest(self, *, body: _ListingLatest) -> ty.Page[_Summary]:
        res = self.request(
            "POST",
            endpoint="all-latest",
            data=body,
            return_model=ty.Page[self.Summary],  # type: ignore
        )
        return ty.cast(ty.Page[_Summary], res)

    async def all_latest_async(self, *, body: _ListingLatest) -> ty.Page[_Summary]:
        res = await self.request_async(
            "POST",
            endpoint="all-latest",
            data=body,
            return_model=ty.Page[self.Summary],  # type: ignore
        )
        return ty.cast(ty.Page[_Summary], res)  # type: ignore

    def latest(self, *, body: _ReadingLatest) -> _Summary:
        res = self.request(
            "POST",
            endpoint="latest",
            data=body,
            return_model=self.Detail,
        )
        return ty.cast(_Summary, res)

    async def latest_async(self, *, body: _ReadingLatest) -> _Summary:
        res = await self.request_async(
            "POST",
            endpoint="latest",
            data=body,
            return_model=self.Detail,
        )
        return ty.cast(_Summary, res)



"""
This file originates in shared/shared_about.py, and
is copied verbatim into the different projects. It should
not be edited directly, edit the shared version and replicate
it everywhere.
"""

import json
from pathlib import Path

from .version.read_version import read_version

PWD = Path(__file__).parent


with (PWD / "about.json").open("r", encoding="utf8") as file_:
    _about_data = json.load(file_)


__name__ = _about_data["name"]
__title__ = _about_data.get("title", "")
__description__ = _about_data.get("description", "")
__summary__ = _about_data.get("summary", "")
__version__ = read_version(PWD / "version")
__prog__ = _about_data.get("prog", "")



from .. import ty
from ..models import (
    BrokerDeleting,
    BrokerDetail,
    BrokerReading,
    BrokerRegistering,
    BrokerRegistrationCreating,
    BrokerRegistrationSummary,
    BrokerSummary,
    BrokerUpdating,
)
from .base import ModelClient


class Broker(
    ModelClient[
        BrokerRegistering,
        BrokerReading,
        BrokerUpdating,
        BrokerDeleting,
        BrokerSummary,
        BrokerDetail,
    ]
):
    Creating = BrokerRegistering
    Reading = BrokerReading
    Updating = BrokerUpdating
    Deleting = BrokerDeleting
    Summary = BrokerSummary
    Detail = BrokerDetail

    class Token(ty.BaseModel):
        token: str

    def create(
        self, data: ty.Optional[BrokerRegistrationCreating] = None, **kwargs: ty.Any
    ) -> BrokerRegistrationSummary:
        data = data if data else BrokerRegistrationCreating(**kwargs)
        res = self.request(
            "POST",
            endpoint="create",
            data=data,
            return_model=BrokerRegistrationSummary,
        )
        return ty.cast(BrokerRegistrationSummary, res)

    async def create_async(
        self, data: ty.Optional[BrokerRegistrationCreating] = None, **kwargs: ty.Any
    ) -> BrokerRegistrationSummary:
        data = data if data else BrokerRegistrationCreating(**kwargs)
        res = await self.request_async(
            "POST",
            endpoint="create",
            data=data,
            return_model=BrokerRegistrationSummary,
        )
        return ty.cast(BrokerRegistrationSummary, res)

    def register(
        self, data: ty.Optional[BrokerRegistering] = None, **kwargs: ty.Any
    ) -> BrokerDetail:
        data = data if data else BrokerRegistering(**kwargs)
        res = self.request(
            "POST",
            endpoint="register",
            data=data,
            return_model=BrokerDetail,
        )
        return ty.cast(BrokerDetail, res)

    async def register_async(
        self, data: ty.Optional[BrokerRegistering] = None, **kwargs: ty.Any
    ) -> BrokerDetail:
        data = data if data else BrokerRegistering(**kwargs)
        res = await self.request_async(
            "POST",
            endpoint="register",
            data=data,
            return_model=BrokerDetail,
        )
        return ty.cast(BrokerDetail, res)



from ..models import (
    AssetCreating,
    AssetDeleting,
    AssetDetail,
    AssetListingLatest,
    AssetReading,
    AssetReadingLatest,
    AssetSummary,
    AssetUpdating,
)
from .base import VersionedModelClient


class Asset(
    VersionedModelClient[
        AssetCreating,
        AssetReading,
        AssetUpdating,
        AssetDeleting,
        AssetSummary,
        AssetDetail,
        AssetReadingLatest,
        AssetListingLatest,
    ]
):
    Creating = AssetCreating
    Reading = AssetReading
    Updating = AssetUpdating
    Deleting = AssetDeleting
    Summary = AssetSummary
    Detail = AssetDetail
    ReadingLatest = AssetReadingLatest
    ListingLatest = AssetListingLatest



from functools import cached_property
from pathlib import Path

from pydantic import BaseModel

from . import ty
from .util import APP_NAME, get_app_dir


def global_config_dir() -> Path:
    return Path(get_app_dir(APP_NAME))


class RootConfig:
    def __init__(self, *, config_dir: Path) -> None:
        self.config_dir = config_dir

    @cached_property
    def secrets_path(self) -> Path:
        return self.config_dir / "secrets.json"

    @cached_property
    def saved_settings_path(self) -> Path:
        return self.config_dir / "saved-defaults.json"


class SavedSettings(BaseModel):
    broker_host: ty.Optional[str]
    project: ty.Optional[ty.UUID]
    task: ty.Optional[ty.UUID]
    action: ty.Optional[ty.UUID]
    pam_host: ty.Optional[str]

    @classmethod
    def blank(cls) -> "SavedSettings":
        return cls(
            broker_host=None, project=None, task=None, action=None, pam_host=None
        )

    @classmethod
    def from_file(cls, path: Path, must_exist: bool = False) -> "SavedSettings":
        try:
            return cls.parse_file(path)
        except FileNotFoundError:
            if not must_exist:
                return cls.blank()
            else:
                raise

    def reset_defaults(self) -> None:
        """Reset defaut project/task/action, e.g. on host changes."""
        self.project = None
        self.task = None
        self.action = None

    def to_json(self) -> ty.JSONableDict:
        data = {}
        for key, value in self.dict().items():
            # UUIDs are not JSON-seriazliable by default
            data[key] = str(value) if isinstance(value, ty.UUID) else value
        return data

    def save(
        self,
        path: Path,
    ) -> None:
        path.parent.mkdir(parents=True, exist_ok=True)
        path.write_text(self.json(), encoding="utf-8")

    @ty.overload
    def get(self, field: ty.Literal["broker_host", "pam_host"]) -> str:
        ...

    @ty.overload
    def get(self, field: ty.Literal["project", "task", "action"]) -> str:
        ...

    def get(
        self, field: ty.Literal["broker_host", "pam_host", "project", "task", "action"]
    ) -> ty.Union[str, ty.UUID]:
        return getattr(self, field)

    @ty.overload
    def update(
        self,
        field: ty.Literal["broker_host", "pam_host"],
        value: ty.Optional[str] = None,
    ) -> str:
        ...

    @ty.overload
    def update(
        self,
        field: ty.Literal["project", "task", "action"],
        value: ty.Optional[ty.UUID] = None,
    ) -> ty.UUID:
        ...

    def update(
        self,
        field: ty.Literal["broker_host", "pam_host", "project", "task", "action"],
        value: ty.Optional[ty.Union[str, ty.UUID]] = None,
    ) -> ty.Union[str, ty.UUID]:
        old_value = getattr(self, field)
        setattr(self, field, value)
        if old_value != value and field in ["broker_host", "pam_host"]:
            self.reset_defaults()
        return getattr(self, field)



import sys

from radicli import (
    DEFAULT_PLACEHOLDER,
    Arg,
    ArgparseArg,
    ArgumentParser,
    ConverterType,
    get_arg,
)
from wasabi import msg

from .. import ty
from ..auth import AuthState
from ..cli import CONVERTERS, cli
from ..config import SavedSettings
from ..errors import CLIError, ProdigyTeamsErrors
from ..messages import Messages
from ..prodigy_teams_broker_sdk.models import JobID, JobType
from ..prodigy_teams_pam_sdk import Client
from ..prodigy_teams_pam_sdk.models import (
    ActionCreating,
    ObjectValidation,
    RecipeDetail,
    RecipeListingLatest,
    RecipePlanCreating,
    TaskCreating,
)
from ..prodigy_teams_pam_sdk.recipe_utils import (
    Field,
    RecipeSchema,
    get_field_cli_args,
    get_field_optional,
    parse_plan,
)
from ..query import resolve_project_id
from ..ui import RecipeHelpFormatter, get_recipe_desc
from ._state import get_auth_state, get_root_cfg


def get_converter(arg_type: ty.Type) -> ty.Optional[ConverterType]:
    return CONVERTERS.get(arg_type, None)


def request_recipes(
    auth: AuthState, is_action: ty.Optional[bool] = None
) -> ty.Dict[str, RecipeDetail]:
    """Fetch all Recipes from Prodigy Teams
    auth (AuthState): CLI Auth object to handle auth to the Prodigy Teams API
    RETURNS (Tuple[Dict[str, RecipeDetail], Dict[str, RecipeDetail]]): Tuple of task_schemas, recipe_schemas
    """
    client = auth.client
    if client is None:
        return {}
    with msg.loading(Messages.T020):
        recipes = auth.retry(client.recipe.all_latest)(
            body=RecipeListingLatest(org_id=auth.org_id, broker_id=auth.broker_id)
        )
    schemas = {}
    for recipe in recipes.items:
        if is_action is None or recipe.is_action == is_action:
            schemas[recipe.name] = recipe
    return schemas


def create_from_recipe(
    recipe: RecipeDetail,
    args: ty.List[str],
    *,
    command: str,
    show_help: bool,
    exists_ok: bool = False,
) -> ty.Tuple[ty.UUID, RecipePlanCreating]:
    try:
        task_or_action_name, project_id, plan = _parse_plan(
            recipe,
            args,
            parser_prog=f"{cli.prog} {command} create {recipe.name.lower()}",
            show_help=show_help,
            get_project_id=_get_default_project_id,
        )
    except ShowHelp as e:
        print(e.help_text)
        sys.exit(1)
    auth = get_auth_state()
    object_validation = _get_object_validation(plan)
    if object_validation.errors:
        errors = ["\n".join(err) for err in object_validation.errors.values()]
        noun = "action" if recipe.is_action else "task"
        err = Messages.E044.format(noun=noun, name=task_or_action_name)
        raise CLIError(err, "\n".join(errors))
    _validate_plan(recipe, plan, object_validation)
    if recipe.is_action:
        job_id = _create_action(
            auth.client, task_or_action_name, project_id, plan, exists_ok=exists_ok
        )
    else:
        job_id = _create_task(
            auth.client, task_or_action_name, project_id, plan, exists_ok=exists_ok
        )
    return job_id.id, plan


def _parse_plan(
    recipe: RecipeDetail,
    args: ty.List[str],
    *,
    parser_prog: str,
    show_help: bool,
    get_project_id: ty.Callable[[], ty.Optional[ty.UUID]],
) -> ty.Tuple[str, ty.UUID, RecipePlanCreating]:
    # This function works hard to be pure and not rely on
    # external state. This way we'll be able to work with it in
    # the compatibility tests more easily.
    noun = "action" if recipe.is_action else "task"
    ap_args = _get_argparse_args(recipe)
    parser = _get_argparse_parser(parser_prog, recipe, ap_args)
    if not args or show_help:
        raise ShowHelp(parser.format_help())
    task_or_action_name, project_id, worker_class, plan_args = _parse_cli_args(
        parser, args
    )
    missing = _check_missing(recipe, ap_args, plan_args)
    if missing:
        err = Messages.E001.format(noun=noun, name=task_or_action_name)
        raise CLIError(err, Messages.E043.format(args=", ".join(missing)))
    if project_id is None:
        project_id = get_project_id()
        if project_id is None:
            raise CLIError(Messages.E042.format(noun=noun, name=task_or_action_name))
    plan = RecipePlanCreating(
        recipe_id=recipe.id,
        project_id=project_id,
        args=plan_args,
        worker_class=worker_class,
    )
    return task_or_action_name, project_id, plan


def _get_default_project_id() -> ty.Optional[ty.UUID]:
    root_cfg = get_root_cfg()
    project_id = SavedSettings.from_file(root_cfg.saved_settings_path).project
    if project_id is not None:
        project_id = resolve_project_id(project_id)
    return project_id


def _parse_cli_args(
    parser: ArgumentParser,
    args: ty.List[str],
) -> ty.Tuple[str, ty.Optional[ty.UUID], ty.Optional[str], ty.Dict[str, Arg]]:
    parsed = vars(parser.parse_args(args))
    task_or_action_name = parsed.pop("name")
    assert isinstance(task_or_action_name, str)
    project_id = parsed.pop("project_id")
    if project_id is not None:
        project_id = ty.UUID(project_id)
    worker_class = parsed.pop("worker_class", None)
    return (task_or_action_name, project_id, worker_class, parsed)


def _check_missing(recipe: RecipeDetail, ap_args, parsed) -> ty.List[str]:
    optional = get_field_optional(recipe.form_schema)
    missing = []
    # We're iterating over the original args here because that gives us access
    # to the option names in the right CLI syntax
    for arg in ap_args:
        required = not optional.get(arg.id, True)
        if required and arg.id not in parsed:
            missing.append(arg.arg.option or arg.id)
    return missing


def _get_argparse_args(recipe: RecipeDetail) -> ty.List[ArgparseArg]:
    noun = "action" if recipe.is_action else "task"
    ap_args: ty.List[ArgparseArg] = []
    ap_args.append(get_arg("name", Arg(help=Messages.name.format(noun=noun)), str))
    for field in recipe.form_schema.fields:
        ap_args.extend(
            get_field_cli_args(
                ty.cast(Field, field), recipe.form_schema.cli_names or {}
            )
        )
    return ap_args


def _get_argparse_parser(
    prog: str, recipe: RecipeDetail, ap_args: ty.List[ArgparseArg]
) -> ArgumentParser:
    noun = "action" if recipe.is_action else "task"
    # To get argument parsing and consistent help docs out-of-the-box we create
    # a dummy parser with radicli's modified ArgumentParser (argparse subclass)
    p = ArgumentParser(
        prog=prog,
        description=get_recipe_desc(recipe),
        formatter_class=RecipeHelpFormatter,
    )
    for arg in ap_args:
        a_args, a_kwargs = arg.to_argparse()
        # We don't want defaults to creep in here, we want that to be
        # introduced at the last moment by the actual function.
        if "default" in a_kwargs:
            a_kwargs["default"] = DEFAULT_PLACEHOLDER
        p.add_argument(*a_args, **a_kwargs)
    # Add default arguments to all recipe CLIs
    pid_arg = Arg("--project-id", help=Messages.recipe_project_id.format(noun=noun))
    pid = get_arg("project_id", pid_arg, str, default=None)
    pid_args, pid_kwargs = pid.to_argparse()
    p.add_argument(*pid_args, **pid_kwargs)
    worker_class_arg = Arg(
        "--worker-class",
        help=Messages.recipe_worker_class.format(noun=noun),
    )
    worker_class = get_arg("worker_class", worker_class_arg, str, default=None)
    worker_class_args, worker_class_kwargs = worker_class.to_argparse()
    p.add_argument(*worker_class_args, **worker_class_kwargs)
    return p


def _get_object_validation(plan: RecipePlanCreating) -> ObjectValidation:
    auth = get_auth_state()
    object_validation = auth.client.recipeplan.validate_objects(plan)
    return object_validation


def _validate_plan(
    recipe: RecipeDetail,
    plan: RecipePlanCreating,
    object_validation: ObjectValidation,
) -> None:
    # Building a dummy objects map for parse_plan – the values don't matter,
    # since we're also only using a dummy fuction for create_custom_type
    objects_map = {"asset": {}, "dataset": {}}
    for field_type, objects in object_validation.existing.items():
        objects_map[field_type] = {arg: {} for arg in objects}
    for field_type, objects in object_validation.to_create.items():
        objects_map[field_type] = {arg: {} for arg in objects}
    parse_plan(
        ty.cast(RecipeSchema, recipe.form_schema),
        plan.args,
        objects_map,
        lambda x: lambda *a, **b: x,
    )


def _create_task(
    client: Client,
    name: str,
    project_id: ty.UUID,
    plan: RecipePlanCreating,
    exists_ok: bool = False,
) -> JobID:
    body = TaskCreating(name=name, project_id=project_id, plan=plan)
    try:
        task = client.task.create(body)
    except ProdigyTeamsErrors.ProjectNotFound:
        err = Messages.E013.format(name=project_id)
        raise CLIError(err)
    except ProdigyTeamsErrors.TaskExists:
        if exists_ok:
            msg.info(Messages.T001.format(noun="task", name=name))
            task = client.task.read(name=name, project_id=project_id)
            return JobID(id=task.id, job_type=JobType.task)
        err = Messages.E014.format(noun="task", name=name, project=project_id)
        raise CLIError(err)
    except ProdigyTeamsErrors.TaskInvalid:
        raise CLIError(Messages.E004.format(noun="task", name=name))
    return JobID(id=task.id, job_type=JobType.task)


def _create_action(
    client: Client,
    name: str,
    project_id: ty.UUID,
    plan: RecipePlanCreating,
    exists_ok: bool = False,
) -> JobID:
    body = ActionCreating(name=name, project_id=project_id, evaluation="", plan=plan)
    try:
        action = client.action.create(body)
    except ProdigyTeamsErrors.ProjectNotFound:
        err = Messages.E013.format(name=project_id)
        raise CLIError(err)
    except ProdigyTeamsErrors.ActionExists:
        if exists_ok:
            msg.info(Messages.T001.format(noun="action", name=name))
            action = client.action.read(name=name, project_id=project_id)
            return JobID(id=action.id, job_type=JobType.action)
        err = Messages.E014.format(noun="action", name=name, project=project_id)
        raise CLIError(err)
    except ProdigyTeamsErrors.ActionInvalid:
        raise CLIError(Messages.E004.format(noun="action", name=name))
    return JobID(id=action.id, job_type=JobType.action)


class ShowHelp(ValueError):
    def __init__(self, help_text: str):
        self.help_text = help_text



import builtins
from pathlib import Path

import srsly
from radicli import Arg
from wasabi import msg

from .. import ty
from ..cli import cli
from ..errors import CLIError, ProdigyTeamsErrors
from ..messages import Messages
from ..prodigy_teams_pam_sdk.models import DatasetDetail, DatasetSummary
from ..query import collect_from_pages, resolve_dataset, resolve_dataset_id
from ..ui import PageProgress, print_info_table, print_table_with_select
from ._state import get_auth_state


@cli.subcommand(
    "datasets",
    "list",
    # fmt: off
    select=Arg("--select", help=Messages.select.format(opts=list(DatasetSummary.__fields__))),
    as_json=Arg("--json", help=Messages.as_json),
    # fmt: on
)
def list(
    select: ty.List[str] = ["id", "name", "kind"], as_json: bool = False
) -> ty.Sequence[DatasetSummary]:
    """List all Datasets"""
    client = get_auth_state().client
    res = client.dataset.all()
    print_table_with_select(res.items, select=select, as_json=as_json)
    return res.items


@cli.subcommand(
    "datasets",
    "info",
    name_or_id=Arg(help=Messages.name_or_id.format(noun="dataset")),
    cluster_id=Arg(help=Messages.cluster_id.format(noun="dataset")),
    select=Arg(
        "--select",
        help=Messages.select.format(opts=builtins.list(DatasetDetail.__fields__)),
    ),
    as_json=Arg("--json", help=Messages.as_json),
)
def info(
    name_or_id: ty.StrOrUUID,
    cluster_id: ty.Optional[ty.UUID] = None,
    select: ty.Optional[ty.List[str]] = None,
    as_json: bool = False,
) -> DatasetDetail:
    """Get detailed info for a Dataset"""
    res = resolve_dataset(name_or_id, broker_id=cluster_id)
    print_info_table(res, as_json=as_json, select=select)
    return res


@cli.subcommand(
    "datasets",
    "create",
    name=Arg(help=Messages.name.format(noun="dataset")),
    kind=Arg("--kind", help=Messages.kind.format(noun="dataset")),
    exists_ok=Arg("--exists-ok", help=Messages.exists_ok),
)
def create(
    name: str, kind: str, exists_ok: bool = False
) -> ty.Union[DatasetSummary, DatasetDetail, None]:
    """Create a new dataset"""
    auth = get_auth_state()
    client = auth.client
    broker_id = auth.broker_id
    try:
        res = client.dataset.create(name=name, kind=kind, broker_id=broker_id)
    except ProdigyTeamsErrors.DatasetExists:
        if exists_ok:
            msg.info(Messages.T001.format(noun="dataset", name=name))
            return None
        raise CLIError(Messages.E002.format(noun="dataset", name=name))
    except ProdigyTeamsErrors.DatasetInvalid:
        raise CLIError(Messages.E004.format(noun="dataset", name=name))
    except ProdigyTeamsErrors.DatasetForbiddenCreate:
        raise CLIError(Messages.E003.format(noun="dataset", name=name))
    msg.divider("Dataset")
    msg.table(res.dict())
    return res


@cli.subcommand(
    "datasets",
    "delete",
    name_or_id=Arg(help=Messages.name_or_id.format(noun="dataset")),
    cluster_id=Arg(help=Messages.cluster_id.format(noun="dataset")),
)
def delete(
    name_or_id: ty.StrOrUUID,
    cluster_id: ty.Optional[ty.UUID] = None,
) -> ty.UUID:
    """Delete a dataset"""
    dataset_id = resolve_dataset_id(name_or_id, broker_id=cluster_id)
    auth = get_auth_state()
    try:
        auth.client.dataset.delete(id=dataset_id)
    except (
        ProdigyTeamsErrors.ProjectForbiddenDelete,
        ProdigyTeamsErrors.ProjectNotFound,
    ):
        raise CLIError(Messages.E006.format(noun="dataset", name=name_or_id))
    else:
        msg.good(Messages.T003.format(noun="dataset", name=name_or_id))
    return dataset_id


@cli.subcommand(
    "datasets",
    "export",
    name_or_id=Arg(help=Messages.name_or_id.format(noun="dataset")),
    output_file=Arg("--output", "-o", help=Messages.export_output),
)
def export(
    name_or_id: ty.StrOrUUID,
    output_file: ty.Union[str, Path] = "-",
) -> None:
    """Export all the examples from a dataset and save it in the designated file as
    JSONL (newline-delimited JSON).
    """
    auth = get_auth_state()
    cluster_id = auth.broker_id

    dataset = resolve_dataset(name_or_id, broker_id=cluster_id)

    def get_examples_page(page: int):
        return auth.broker_client.datasets.read_examples(
            [dataset.name],
            page=page,
        )

    with PageProgress(collect_from_pages(get_examples_page)) as pages:
        srsly.write_jsonl(
            output_file,
            (example.content for page in pages for example in page.items),
            append_new_line=False,
        )



from radicli import Arg
from wasabi import msg

from .. import ty
from ..auth import AuthState
from ..cli import cli
from ..config import SavedSettings, global_config_dir
from ..errors import CLIError, ProdigyTeamsError
from ..messages import Messages
from ..ui import print_as_json
from ._state import get_auth_state, get_root_cfg


@cli.command(
    "login",
    no_cluster=Arg("--no-cluster", help=Messages.no_cluster),
)
def login(no_cluster: bool = False) -> None:
    """
    Log in to your Prodigy Teams account. You normally don't need to call this
    manually. It will automatically authenticate when needed.
    """
    auth = get_auth_state()
    auth._ensure_readable_secrets()
    auth.get_id_token(force_refresh=True)
    auth.get_api_token(force_refresh=True)
    if not no_cluster:
        try:
            auth.get_broker_token(force_refresh=True)
        except ProdigyTeamsError as e:
            err = Messages.E116.format(command=f"{cli.prog} login --no-cluster")
            raise CLIError(err, e)
    msg.good(Messages.T012)


@cli.command("info", field=Arg(help=Messages.select_field))
def info(
    field: ty.Optional[ty.Literal["config-dir", "code", "defaults"]] = None
) -> ty.Any:
    """Print information about the CLI"""
    settings = SavedSettings.from_file(get_root_cfg().saved_settings_path)
    info = {
        "config-dir": str(global_config_dir().absolute()),
        "code": __file__,
        "defaults": settings.to_json(),
    }
    if field:
        print(info[field])
        return info[field]
    else:
        print_as_json(info)
        return info


@cli.command(
    "get-auth-token",
    token_type=Arg(help="The token type"),
)
def get_auth_token(
    token_type: ty.Optional[ty.Literal["api", "cluster", "id", "ci"]] = None
) -> AuthState:
    """
    Return an auth token for the current user
    """
    auth = get_auth_state()
    if token_type == "api":
        print(auth.get_api_token().access_token)
    elif token_type == "cluster":
        print(auth.get_broker_token().access_token)
    elif token_type == "id":
        print(auth.get_id_token())
    elif token_type == "ci":
        # this uses a dev endpoint because we don't want to give CI
        # a full access token. Once we can create and revoke limited
        # tokens, remove this.
        response = auth.client._sync_client.post(
            "/api/v1/dev/create-ci-token",
            json={},
        )
        response.raise_for_status()
        print(response.json()["access_token"])
    else:
        raise CLIError(Messages.E117.format(token_type=token_type))
    return auth
